#!/usr/bin/env python
#
#    Copyright (C) 2010-2017 PyTRiP98 Developers.
#
#    This file is part of PyTRiP98.
#
#    PyTRiP98 is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyTRiP98 is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyTRiP98.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Script for generating DVH or other volume histograms.
"""
import sys
import logging
import argparse

import pytrip as pt
from pytrip.volhist import VolHist

import matplotlib

logger = logging.getLogger(__name__)


def main(args=sys.argv[1:]):
    """ Main function for dvhplot.py
    """

    # parse arguments
    parser = argparse.ArgumentParser()
    parser.add_argument("cube", help="path to input cube. May also be a .dos or dosemlet.dos cube", type=str)
    parser.add_argument("vdx", help="path to .vdx file holding the structures", type=str)
    parser.add_argument("rois", nargs="?", help="comma-seperated list for ROIs to be analyzed. If not set, print list.",
                        type=str, default=None)
    parser.add_argument("-d", "--dose", type=float, dest='dose', metavar='dose',
                        help="target dose in [Gy] (if target_dose is unavailable in cube)", default=None)
    parser.add_argument("-o", "--output", type=str, dest='output', metavar='filename',
                        help="don't open GUI, save figure to <filename> instead.", default=None)
    parser.add_argument("-t", "--tofile", type=str, dest='tofile', metavar='filename',
                        help="save histogram data to <filename>.", default=None)
    parser.add_argument("-l", "--legend", dest='legend', default=False, action='store_true',
                        help="print legend box")
    parser.add_argument("-v", "--verbosity", action='count', help="increase output verbosity", default=0)
    parser.add_argument('-V', '--version', action='version', version=pt.__version__)
    parsed_args = parser.parse_args(args)

    if parsed_args.verbosity == 1:
        logging.basicConfig(level=logging.INFO)
    elif parsed_args.verbosity > 1:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig()

    path_cube = parsed_args.cube
    path_vdx = parsed_args.vdx
    rois_arg = parsed_args.rois
    dose = parsed_args.dose
    legend = parsed_args.legend
    outfile = parsed_args.output
    tofile = parsed_args.tofile

    # there are some cases when this script is run on systems without DISPLAY variable being set
    # in such case matplotlib backend has to be explicitly specified
    # despite the fact interleaving imports with code lines is discouraged, we do it here
    # as it depends on the users options
    if outfile:
        matplotlib.use('Agg')
    import matplotlib.pyplot as plt

    d = pt.DosCube()
    d.read(path_cube)

    v = pt.VdxCube(d)
    v.read(path_vdx)

    vois = v.voi_names()

    if not rois_arg:
        print("Available ROIs:")
        for voi in vois:
            print("'{:s}'".format(voi))
        return

    rois = rois_arg.split(",")

    for roi in rois:
        voi = v.get_voi_by_name(roi)
        vh = VolHist(d, voi, target_dose=dose)

        plt.xlabel(vh.xlabel)
        plt.ylabel(vh.ylabel)
        plt.plot(vh.x, vh.y, label=vh.name)

    plt.ylabel("Volume [%]")
    plt.grid(True)
    if legend:
        plt.legend()

    if tofile:
        logger.info("Write {}".format(tofile))
        f = open(tofile, "w+")
        for _x, _y in zip(vh.x, vh.y):
            f.write("{:.3f} {:.3f}\n".format(_x, _y))
        f.close()

    if outfile:
        plt.savefig(outfile)
    else:
        plt.show()


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



#
#    Copyright (C) 2010-2017 PyTRiP98 Developers.
#
#    This file is part of PyTRiP98.
#
#    PyTRiP98 is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyTRiP98 is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyTRiP98.  If not, see <http://www.gnu.org/licenses/>.
#
"""
The models module provides functions for calculating cell survival and RBE.
"""

from pytrip.models.proton import rbe_carabe, rbe_wedenberg, rbe_mcnamara
from pytrip.models.rcr import rbe_rcr, sf_rcr, oer_rcr, oer_po2_rcr
from pytrip.models.tcp import tcp_voi
from pytrip.models.extra import rbe_from_sf, lq

__all__ = ['rbe_carabe', 'rbe_wedenberg', 'rbe_mcnamara',
           'rbe_rcr', 'sf_rcr', 'oer_rcr', 'oer_po2_rcr',
           'tcp_voi',
           'rbe_from_sf', 'lq']



#
#    Copyright (C) 2010-2017 PyTRiP98 Developers.
#
#    This file is part of PyTRiP98.
#
#    PyTRiP98 is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyTRiP98 is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyTRiP98.  If not, see <http://www.gnu.org/licenses/>.
#
"""
The tripexecuter module provides functions for executing TRiP98 locally or remotely.
"""

from pytrip.tripexecuter.field import Field
from pytrip.tripexecuter.execute import Execute
from pytrip.tripexecuter.plan import Plan
from pytrip.tripexecuter.projectile import Projectile
from pytrip.tripexecuter.kernel import KernelModel
from pytrip.tripexecuter.execparser import ExecParser

# from https://docs.python.org/3/tutorial/modules.html
# if a package's __init__.py code defines a list named __all__,
# it is taken to be the list of module names that should be imported when from package import * is encountered.
__all__ = ['Field', 'Execute', 'Plan', 'Projectile', 'KernelModel', 'ExecParser']



#
#    Copyright (C) 2010-2018 PyTRiP98 Developers.
#
#    This file is part of PyTRiP98.
#
#    PyTRiP98 is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyTRiP98 is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyTRiP98.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Object holding all data needed for a specific projectile/rifi configuration.
"""

import logging


logger = logging.getLogger(__name__)


class Projectile(object):
    """
    Object holding projectile specific data.
    """

    # list of projectile name - charge and most common isotope.
    _projectile_defaults = {"H": (1, 1),
                            "He": (2, 4),
                            "Li": (3, 7),
                            "C": (6, 12),
                            "O": (8, 16),
                            "Ne": (10, 20),
                            "Ar": (16, 40)}

    z = -1
    a = -1
    iupac = ""  # chemical symbol (parsed by TRiP), "H", "He", "C", "O", "Ne" ....
    name = ""   # cleartext name "Protons", "Carbon-12 ions", "Antiprotons", ...

    def __init__(self, symb, name="", z=0, a=0):
        """
        :param symb: IUPAC symbol "H", "He", "Li", "C" ...
        :param name: free text name for this projectile, i.e. "Protons", "Antiprotons", "C-12"
        :param z: charge of projectile. Default value for symb taken if not specified.
        :param a: nucleon number of proejctile, default value for symb is taken if not specified (12 for C, 4 for He...)
        """

        self.name = name

        if symb in self._projectile_defaults:
            self.z = self._projectile_defaults[symb][0]
            self.a = self._projectile_defaults[symb][1]

        self.iupac = symb

        if z > 0:
            self.z = z

        if a > 0:
            self.a = a

        if self.z < 1:
            logger.warning("No projectile charge was set.")
        if self.a < 1:
            logger.warning("No projectile nucleon number was set.")

    def __str__(self):
        """
        String out handler
        """
        return self._print()

    def _print(self):
        """
        Pretty print all attributes.
        """
        out = "\n"
        out += "   Projectile '{:s}'\n".format(self.name)
        out += "----------------------------------------------------------------------------\n"
        out += "|  Projectile Z                      : {:s}\n".format(str(self.z))
        out += "|  Projectile A                      : {:s}\n".format(str(self.a))
        out += "|  Projectile IUPAC                  : {:s}\n".format(str(self.iupac))
        out += "----------------------------------------------------------------------------\n"
        return out



#
#    Copyright (C) 2010-2017 PyTRiP98 Developers.
#
#    This file is part of PyTRiP98.
#
#    PyTRiP98 is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyTRiP98 is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyTRiP98.  If not, see <http://www.gnu.org/licenses/>.
#
"""
This module provides classes to handle SPC data

SPC binary file format is described here:
http://bio.gsi.de/DOCS/TRiP98/PRO/DOCS/trip98fmtspc.html


"""

import os
from enum import Enum
import logging
import numpy as np


class SPCTagCode(Enum):
    FILETYPE = 1
    FILEVERSION = 2
    FILEDATE = 3
    TARGNAME = 4
    PROJNAME = 5
    BEAM_ENERGY = 6
    PEAK_POSITION = 7
    NORMALIZATION = 8
    NO_OF_DEPTH_STEPS = 9
    DEPTH = 10
    DEPTH_STEP_NORMALIZATION = 11
    NO_OF_PARTICLE_SPECIES = 12
    ZA = 13
    CUMULATED_NO_OF_FRAGMENTS = 14
    NC = 15
    NO_OF_ENERGY_BINS = 16
    ENERGY_BIN_VALUES = 17
    ENERGY_BIN_POINTER = 18
    SPECTRUM_BIN_VALUES = 19
    CUMULATED_SPECTRUM_BIN_VALUES = 20


class SPCCollection(object):
    def __init__(self, dirname):
        self.dirname = dirname
        self.data = {}  # key - energy, value - SPC object

    def read(self):
        files = os.listdir(self.dirname)
        for file in files:
            if file.endswith(".spc"):
                energy = int(file.split('.')[-2][3:]) / 100
                print("energy", energy)
                self.data[energy] = SPC(os.path.join(self.dirname, file))
                self.data[energy].read_spc()

    def write(self):
        if not os.path.exists(self.dirname):
            os.makedirs(self.dirname)
        for spc_object in self.data.values():
            fname = "{pp:s}.{tt:s}.{uuu:s}{eeeee:05d}.spc".format(
                pp=spc_object.projname,
                tt=spc_object.targname,
                uuu="MeV",
                eeeee=int(100 * spc_object.energy)
            )
            print(fname)
            spc_object.write_spc(os.path.join(self.dirname, fname))


class SPC(object):
    def __init__(self, filename):
        self.filename = filename
        self.data = []

    def read_spc(self):
        logging.info("Reading {}".format(self.filename))
        self.read_data()

    def write_spc(self, filename=None):
        fname = filename
        if fname is None:
            fname = "{pp:s}.{tt:s}.{uuu:s}{eeeee:05d}.spc".format(
                pp=self.projname,
                tt=self.targname,
                uuu="MeV",
                eeeee=int(100 * self.energy)
            )
        fd = open(fname, "wb")

        # filetype 1
        # <filetype> is an 80-byte ASCII character string starting with "SPCM" or "SPCI",
        # specifying big ("Motorola") or little ("Intel") endian byte order, respectively.
        tag = Tag(fd)
        tag.code = SPCTagCode.FILETYPE.value
        tag.size = 80
        tag.endian = self.endian
        tag.set_tag()
        fd.write(np.asarray(self.filetype, dtype="{:s}a{:d}".format(tag._ste, tag.size)))

        # fileversion 2
        # <fileversion> is an 80-byte ASCII character string specifying the file format version as yyyymmdd.
        # 19980704 is the "classic" format (fixed energy range),
        # whereas 20020107 is reserved for future possible variable energy range.
        tag = Tag(fd)
        tag.code = SPCTagCode.FILEVERSION.value
        tag.size = 80
        tag.endian = self.endian
        tag.set_tag()
        fd.write(np.asarray(self.fileversion, dtype="{:s}a{:d}".format(tag._ste, tag.size)))

        # filedate 3
        # <filedate> is an 80-byte ASCII character string with the file creation date
        # as returned by the ctime() function. (<dow> <mmm> <dd> <hh>:<mm>:<ss> <yyyy>)
        tag = Tag(fd)
        tag.code = SPCTagCode.FILEDATE.value
        tag.size = 80
        tag.endian = self.endian
        tag.set_tag()
        fd.write(np.asarray(self.filedate, dtype="{:s}a{:d}".format(tag._ste, tag.size)))

        # targname 4, projname 5
        # <targname> and <projname> are the names of target ("H2O") and projectile ("12C6"), respectively.
        # Since both can have any length, they are padded to the right with binary zeroes
        # up to the next 8-byte boundary.
        tag = Tag(fd)
        tag.code = SPCTagCode.TARGNAME.value
        tag.size = 8  # TODO to be padded
        tag.endian = self.endian
        tag.set_tag()
        fd.write(np.asarray(self.targname, dtype="{:s}a{:d}".format(tag._ste, tag.size)))

        tag = Tag(fd)
        tag.code = SPCTagCode.PROJNAME.value
        tag.size = 8  # TODO to be padded
        tag.endian = self.endian
        tag.set_tag()
        fd.write(np.asarray(self.projname, dtype="{:s}a{:d}".format(tag._ste, tag.size)))

        # no 6 double; beam energy [MeV/u]
        tag = Tag(fd)
        tag.code = SPCTagCode.BEAM_ENERGY.value
        tag.size = 8
        tag.endian = self.endian
        tag.set_tag()
        fd.write(np.asarray(self.energy, dtype="{:s}f{:d}".format(tag._ste, tag.size)))

        # no 7, double; peak position [g/cm**2]
        tag = Tag(fd)
        tag.code = SPCTagCode.PEAK_POSITION.value
        tag.size = 8
        tag.endian = self.endian
        tag.set_tag()
        fd.write(np.asarray(self.peakpos, dtype="{:s}f{:d}".format(tag._ste, tag.size)))

        # no 8, double; normalization, usually =1
        tag = Tag(fd)
        tag.code = SPCTagCode.NORMALIZATION.value
        tag.size = 8
        tag.endian = self.endian
        tag.set_tag()
        fd.write(np.asarray(self.norm, dtype="{:s}f{:d}".format(tag._ste, tag.size)))

        # no 9, 8-byte unsigned integer; number of depth steps.
        tag = Tag(fd)
        tag.code = SPCTagCode.NO_OF_DEPTH_STEPS.value
        tag.size = 8
        tag.endian = self.endian
        tag.set_tag()
        fd.write(np.asarray(self.ndsteps, dtype="{:s}i{:d}".format(tag._ste, tag.size)))

        for dblock in self.data:
            # no 10, double; depth [g/cm**2]
            tag = Tag(fd)
            tag.code = SPCTagCode.DEPTH.value
            tag.size = 8
            tag.endian = self.endian
            tag.set_tag()
            fd.write(np.asarray(dblock.depth, dtype="{:s}f{:d}".format(tag._ste, tag.size)))

            # no 11, double; normalization for this depth step, usually =1.
            tag = Tag(fd)
            tag.code = SPCTagCode.DEPTH_STEP_NORMALIZATION.value
            tag.size = 8
            tag.endian = self.endian
            tag.set_tag()
            fd.write(np.asarray(dblock.dsnorm, dtype="{:s}f{:d}".format(tag._ste, tag.size)))

            # no 12, 8-byte unsigned integer; number of particle species.
            tag = Tag(fd)
            tag.code = SPCTagCode.NO_OF_PARTICLE_SPECIES.value
            tag.size = 8
            tag.endian = self.endian
            tag.set_tag()
            fd.write(np.asarray(dblock.nparts, dtype="{:s}u{:d}".format(tag._ste, tag.size)))

            ebins = {}

            for l, specie in enumerate(dblock.species):
                # no 13, double Z, double A, long Z, long A;
                tag = Tag(fd)
                tag.code = SPCTagCode.ZA.value
                tag.size = 24
                tag.endian = self.endian
                tag.set_tag()
                fd.write(np.asarray([specie.z, specie.a], dtype="{:s}f8".format(tag._ste)))
                fd.write(np.asarray([specie.lz, specie.la], dtype="{:s}u4".format(tag._ste)))

                # no 14, double
                # The scalar Cum value is the cumulated number (running sum) of fragments,
                # i.e. the species sum over Cum[nE].
                # This number may exceed 1, since for an incoming primary particle many secondaries are created.
                tag = Tag(fd)
                tag.code = SPCTagCode.CUMULATED_NO_OF_FRAGMENTS.value
                tag.size = 8
                tag.endian = self.endian
                tag.set_tag()
                fd.write(np.asarray([specie.dscum], dtype="{:s}f{:d}".format(tag._ste, tag.size)))

                # no 15, 8-byte unsigned integer;
                # <nC> is reserved for later use, so that lateral scattering for each fragment can be included.
                # At present nC=0.
                tag = Tag(fd)
                tag.code = SPCTagCode.NC.value
                tag.size = 8
                tag.endian = self.endian
                tag.set_tag()
                fd.write(np.asarray([specie.nc], dtype="{:s}u{:d}".format(tag._ste, tag.size)))

                # no 16, 8-byte unsigned integer;
                # 8-byte unsigned integer; number of energy bins for this species
                tag = Tag(fd)
                tag.code = SPCTagCode.NO_OF_ENERGY_BINS.value
                tag.size = 8
                tag.endian = self.endian
                tag.set_tag()
                fd.write(np.asarray([specie.ne], dtype="{:s}u{:d}".format(tag._ste, tag.size)))

                # If a species data block is flagged as EREFCOPY, its energy bins are not stored,
                # but rather a reference index <lSRef> (0..<nS>-1) to a species with identical energy binning.
                # This helps to reduce space requirements, since fragment spectra may share the same energy bins.
                # If the energy bins for the species under consideration are not a reference copy,
                # <nE>+1 double precision bin border values are stored.
                compatible_ebins_index = [i for i in ebins if np.array_equal(specie.ebindata, ebins[i])]
                if not compatible_ebins_index:
                    ebins[l] = specie.ebindata

                    # no 17, double; energy bin values
                    tag = Tag(fd)
                    tag.code = SPCTagCode.ENERGY_BIN_VALUES.value
                    tag.size = 8 * (specie.ne + 1)
                    tag.endian = self.endian
                    tag.set_tag()
                    fd.write(np.asarray(specie.ebindata, dtype="{:s}f8".format(tag._ste)))
                else:
                    # no 18,  8-byte unsigned integer
                    tag = Tag(fd)
                    tag.code = SPCTagCode.ENERGY_BIN_POINTER.value
                    tag.size = 8
                    tag.endian = self.endian
                    tag.set_tag()
                    fd.write(np.asarray([compatible_ebins_index[0]], dtype="{:s}u{:d}".format(tag._ste, tag.size)))

                # no 19,  double; spectrum bin values
                tag = Tag(fd)
                tag.code = SPCTagCode.SPECTRUM_BIN_VALUES.value
                tag.size = 8 * specie.ne
                tag.endian = self.endian
                tag.set_tag()
                fd.write(np.asarray(specie.histdata, dtype="{:s}f8".format(tag._ste)))

                # no 20,  double; running cumulated spectrum bin values
                tag = Tag(fd)
                tag.code = SPCTagCode.CUMULATED_SPECTRUM_BIN_VALUES.value
                tag.size = 8 * (specie.ne + 1)
                tag.endian = self.endian
                tag.set_tag()
                fd.write(np.asarray(specie.rcumdata, dtype="{:s}f8".format(tag._ste)))

        fd.close()

    def read_data(self):
        fd = open(self.filename, "rb")
        t = Tag(fd)
        db = []

        while t.code < SPCTagCode.NO_OF_DEPTH_STEPS.value:
            t.get_tag()
            self.endian = t.endian
            pl = self.get_payload(fd, t)

        for i in range(self.ndsteps):
            # no 10, double; depth [g/cm**2]
            t.get_tag()
            pl = self.get_payload(fd, t)

            # construct the main object
            if t.code == SPCTagCode.DEPTH.value:
                db.append(DBlock())
                db[-1].depth = pl

                # no 11, double; normalization for this depth step, usually =1.
                t.get_tag()
                if t.code == SPCTagCode.DEPTH_STEP_NORMALIZATION.value:
                    pl = self.get_payload(fd, t)
                    db[-1].dsnorm = pl

                # no 12, 8-byte unsigned integer; number of particle species.
                t.get_tag()
                if t.code == SPCTagCode.NO_OF_PARTICLE_SPECIES.value:
                    pl = self.get_payload(fd, t)
                    db[-1].nparts = pl

                for j in range(db[-1].nparts):  # loop over all species
                    # no 13, double Z, double A, long Z, long A;
                    t.get_tag()
                    if t.code == SPCTagCode.ZA.value:
                        pl = self.get_payload(fd, t)
                        db[-1].species.append(SBlock())
                        db[-1].species[-1].z = pl[0]
                        db[-1].species[-1].a = pl[1]
                        db[-1].species[-1].lz = int(pl[2])
                        db[-1].species[-1].la = int(pl[3])

                    # no 14, double
                    # The scalar Cum value is the cumulated number (running sum) of fragments,
                    # i.e. the species sum over Cum[nE].
                    # This number may exceed 1, since for an incoming primary particle many secondaries are created.
                    t.get_tag()
                    if t.code == SPCTagCode.CUMULATED_NO_OF_FRAGMENTS.value:
                        pl = self.get_payload(fd, t)
                        db[-1].species[-1].dscum = pl

                    # no 15, 8-byte unsigned integer;
                    # <nC> is reserved for later use, so that lateral scattering for each fragment can be included.
                    # At present nC=0.
                    t.get_tag()
                    if t.code == SPCTagCode.NC.value:
                        pl = self.get_payload(fd, t)
                        db[-1].species[-1].nc = pl

                    # no 16, 8-byte unsigned integer;
                    # 8-byte unsigned integer; number of energy bins for this species
                    t.get_tag()
                    if t.code == SPCTagCode.NO_OF_ENERGY_BINS.value:
                        pl = self.get_payload(fd, t)
                        db[-1].species[-1].ne = pl

                    # no 17, double; energy bin values
                    t.get_tag()
                    if t.code == SPCTagCode.ENERGY_BIN_VALUES.value:
                        pl = self.get_payload(fd, t)
                        db[-1].species[-1].ebindata = pl
                    # no 18,  8-byte unsigned integer
                    if t.code == SPCTagCode.ENERGY_BIN_POINTER.value:  # both tags will not be present
                        pl = self.get_payload(fd, t)
                        db[-1].species[-1].ebindata = db[0].species[pl].ebindata

                    # no 19,  double; spectrum bin values
                    t.get_tag()
                    if t.code == SPCTagCode.SPECTRUM_BIN_VALUES.value:
                        pl = self.get_payload(fd, t)
                        db[-1].species[-1].histdata = pl

                    # no 20,  double; running cumulated spectrum bin values
                    t.get_tag()
                    if t.code == SPCTagCode.CUMULATED_SPECTRUM_BIN_VALUES.value:
                        pl = self.get_payload(fd, t)
                        db[-1].species[-1].rcumdata = pl

            self.data = db

    def get_payload(self, fd, tag):
        cnt = 1

        if self.endian == 0:
            ste = '<'  # little endian
        else:
            ste = '>'  # big endian

        # filetype 1
        # <filetype> is an 80-byte ASCII character string starting with "SPCM" or "SPCI",
        # specifying big ("Motorola") or little ("Intel") endian byte order, respectively.
        if tag.code == SPCTagCode.FILETYPE.value:
            sdtype = ste + 'a' + str(tag.size)
            payload = np.fromfile(fd, count=cnt, dtype=np.dtype(sdtype))[0].decode("UTF-8")
            self.filetype = payload
            return payload

        # fileversion 2
        # <fileversion> is an 80-byte ASCII character string specifying the file format version as yyyymmdd.
        # 19980704 is the "classic" format (fixed energy range),
        # whereas 20020107 is reserved for future possible variable energy range.
        if tag.code == SPCTagCode.FILEVERSION.value:
            sdtype = ste + 'a' + str(tag.size)
            payload = np.fromfile(fd, count=cnt, dtype=np.dtype(sdtype))[0].decode("UTF-8")
            self.fileversion = payload
            return payload

        # filedate 3
        # <filedate> is an 80-byte ASCII character string with the file creation date
        # as returned by the ctime() function. (<dow> <mmm> <dd> <hh>:<mm>:<ss> <yyyy>)
        if tag.code == SPCTagCode.FILEDATE.value:
            sdtype = ste + 'a' + str(tag.size)
            payload = np.fromfile(fd, count=cnt, dtype=np.dtype(sdtype))[0].decode("UTF-8")
            self.filedate = payload
            return payload

        # targname 4, projname 5
        # <targname> and <projname> are the names of target ("H2O") and projectile ("12C6"), respectively.
        # Since both can have any length, they are padded to the right with binary zeroes
        # up to the next 8-byte boundary.
        if tag.code == SPCTagCode.TARGNAME.value:
            sdtype = ste + 'a' + str(tag.size)
            payload = np.fromfile(fd, count=cnt, dtype=np.dtype(sdtype))[0].decode("UTF-8")
            self.targname = payload
            return payload

        if tag.code == SPCTagCode.PROJNAME.value:
            sdtype = ste + 'a' + str(tag.size)
            payload = np.fromfile(fd, count=cnt, dtype=np.dtype(sdtype))[0].decode("UTF-8")
            self.projname = payload
            return payload

        # no 6 double; beam energy [MeV/u]
        if tag.code == SPCTagCode.BEAM_ENERGY.value:
            sdtype = ste + 'f' + str(tag.size)
            payload = np.fromfile(fd, count=cnt, dtype=np.dtype(sdtype))[0]
            self.energy = payload
            return payload

        # no 7, double; peak position [g/cm**2]
        if tag.code == SPCTagCode.PEAK_POSITION.value:
            sdtype = ste + 'f' + str(tag.size)
            payload = np.fromfile(fd, count=cnt, dtype=np.dtype(sdtype))[0]
            self.peakpos = payload
            return payload

        # no 8, double; normalization, usually =1
        if tag.code == SPCTagCode.NORMALIZATION.value:
            sdtype = ste + 'f' + str(tag.size)
            payload = np.fromfile(fd, count=cnt, dtype=np.dtype(sdtype))[0]
            self.norm = payload
            return payload

        # no 9, 8-byte unsigned integer; number of depth steps.
        if tag.code == SPCTagCode.NO_OF_DEPTH_STEPS.value:
            sdtype = ste + 'u' + str(tag.size)
            payload = np.fromfile(fd, count=cnt, dtype=np.dtype(sdtype))[0]
            self.ndsteps = payload
            return payload

        # no 10, double; depth [g/cm**2]
        if tag.code == SPCTagCode.DEPTH.value:
            sdtype = ste + 'f' + str(tag.size)
            payload = np.fromfile(fd, count=cnt, dtype=np.dtype(sdtype))[0]
            # self.depth = payload
            return payload

        # no 11, double; normalization for this depth step, usually =1.
        if tag.code == SPCTagCode.DEPTH_STEP_NORMALIZATION.value:
            sdtype = ste + 'f' + str(tag.size)
            payload = np.fromfile(fd, count=cnt, dtype=np.dtype(sdtype))[0]
            # self.dsnorm = payload
            return payload

        # no 12, 8-byte unsigned integer; number of particle species.
        if tag.code == SPCTagCode.NO_OF_PARTICLE_SPECIES.value:
            sdtype = ste + 'u' + str(tag.size)
            payload = np.fromfile(fd, count=cnt, dtype=np.dtype(sdtype))[0]
            # self.nparts = payload
            return payload

        # no 13, double Z, double A, long Z, long A;
        if tag.code == SPCTagCode.ZA.value:
            payload = np.zeros(4)

            sdtype = ste + 'f8'
            payload[0] = np.fromfile(fd, count=cnt, dtype=np.dtype(sdtype))
            payload[1] = np.fromfile(fd, count=cnt, dtype=np.dtype(sdtype))

            # todo: fix types
            sdtype = ste + 'u4'
            payload[2] = np.fromfile(fd, count=cnt, dtype=np.dtype(sdtype))
            payload[3] = np.fromfile(fd, count=cnt, dtype=np.dtype(sdtype))

            return payload

        # no 14, double
        # The scalar Cum value is the cumulated number (running sum) of fragments,
        # i.e. the species sum over Cum[nE].
        # This number may exceed 1, since for an incoming primary particle many secondaries are created.
        if tag.code == SPCTagCode.CUMULATED_NO_OF_FRAGMENTS.value:
            sdtype = ste + 'f' + str(tag.size)
            payload = np.fromfile(fd, count=cnt, dtype=np.dtype(sdtype))[0]
            # self.dsnorm = payload
            return payload

        # no 15, 8-byte unsigned integer;
        # <nC> is reserved for later use, so that lateral scattering for each fragment can be included.
        # At present nC=0.
        if tag.code == SPCTagCode.NC.value:
            sdtype = ste + 'u' + str(tag.size)
            payload = np.fromfile(fd, count=cnt, dtype=np.dtype(sdtype))[0]
            # self.nc = payload
            return payload

        # no 16, 8-byte unsigned integer;
        # 8-byte unsigned integer; number of energy bins for this species
        if tag.code == SPCTagCode.NO_OF_ENERGY_BINS.value:
            sdtype = ste + 'u' + str(tag.size)
            payload = np.fromfile(fd, count=cnt, dtype=np.dtype(sdtype))[0]
            # self.ne = payload
            return payload

        # no 17, double; energy bin values
        if tag.code == SPCTagCode.ENERGY_BIN_VALUES.value:
            sdtype = ste + 'f8'
            cnt = tag.size // 8
            payload = np.fromfile(fd, count=cnt, dtype=np.dtype(sdtype))
            # self.ebindata = payload
            return payload

        # no 18,  8-byte unsigned integer
        # EREF: if tag is set, then use copy of ebin from species os stated in the file.
        if tag.code == SPCTagCode.ENERGY_BIN_POINTER.value:
            sdtype = ste + 'u8'
            cnt = tag.size // 8
            payload = np.fromfile(fd, count=cnt, dtype=np.dtype(sdtype))[0]
            # self.eref = payload
            return payload

        # no 19,  double; spectrum bin values
        # H[nE]: spectrum contents divided by the bin width
        if tag.code == SPCTagCode.SPECTRUM_BIN_VALUES.value:
            sdtype = ste + 'f8'
            cnt = tag.size // 8
            payload = np.fromfile(fd, count=cnt, dtype=np.dtype(sdtype))
            # self.histdata = payload
            return payload

        # no 20,  double; running cumulated spectrum bin values
        if tag.code == SPCTagCode.CUMULATED_SPECTRUM_BIN_VALUES.value:
            sdtype = ste + 'f8'
            cnt = tag.size // 8
            payload = np.fromfile(fd, count=cnt, dtype=np.dtype(sdtype))
            # self.cum = payload
            return payload


class DBlock(object):  # the depth block
    def __init__(self):
        self.depth = 0.0
        self.norm = 0.0
        self.nparts = 0
        self.species = []  # list of species, for SBlock class


class SBlock(object):  # particle species block
    def __init__(self):
        self.z = 0.0
        self.a = 0.0
        self.lz = 0
        self.la = 0
        self.nc = 0
        self.ne = 0
        self.ebindata = []
        self.histdata = []
        self.rcumdata = []


class Tag(object):
    def __init__(self, fd):
        self.fd = fd
        self.code = 0
        self.size = 0
        self.endian = -1  # 0 little endian, 1 big endian, -1 unknown
        self._ste = ''
        pass

    def set_tag(self):
        if self.endian == 1:  # big endian (Motorola)
            self._ste = '>'
        else:
            self._ste = '<'  # little endian (Intel)
        self.fd.write(np.asarray([self.code, self.size], dtype=self._ste + "u4"))

    def get_tag(self):
        if self.endian == -1:
            # try Intel first, little endian
            self.endian = 0
            self._ste = '<'  # little endian read
            code = np.fromfile(self.fd, count=1, dtype=np.dtype(self._ste + 'u4'))[0]
            self.fd.seek(-4, 1)  # rewind 4 bytes

            if code < 1 or code > 20:  # if fail Intel, then:
                self.endian = 1  # big endian
                self._ste = '>'  # big endian read
                code = np.fromfile(self.fd, count=1, dtype=np.dtype(self._ste + 'u4'))[0]
                self.fd.seek(-4, 1)  # rewind 4 bytes, retry
                if code < 1 or code > 20:
                    print("Error: bad format in SPC file.")
                    exit()
                else:
                    # print("Found big-endian format.")
                    pass

        self.code = np.fromfile(self.fd, count=1, dtype=np.dtype(self._ste + 'u4'))[0]
        self.size = np.fromfile(self.fd, count=1, dtype=np.dtype(self._ste + 'u4'))[0]



#
#    Copyright (C) 2010-2017 PyTRiP98 Developers.
#
#    This file is part of PyTRiP98.
#
#    PyTRiP98 is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyTRiP98 is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyTRiP98.  If not, see <http://www.gnu.org/licenses/>.
#
"""
This module provides the Cube class, which is used by the CTX, DOS, LET and VDX modules.
A cube is a 3D object holding data, such as CT Hounsfield units, Dose- or LET values.
"""
import os
import re
import sys
import logging
import datetime

import numpy as np

try:
    # as of version 1.0 pydicom package import has beed renamed from dicom to pydicom
    from pydicom import uid
    from pydicom.dataset import Dataset, FileDataset
    _dicom_loaded = True
except ImportError:
    try:
        # fallback to old (<1.0) pydicom package version
        from dicom import UID as uid  # old pydicom had UID instead of uid
        from dicom.dataset import Dataset, FileDataset
        _dicom_loaded = True
    except ImportError:
        _dicom_loaded = False

from pytrip.error import InputError, ModuleNotLoadedError, FileNotFound
from pytrip.util import TRiP98FilePath, TRiP98FileLocator

logger = logging.getLogger(__name__)


class Cube(object):
    """ Top level class for 3-dimensional data cubes used by e.g. DosCube, CtxCube and LETCube.
    Otherwise, this cube class may be used for storing different kinds of data, such as number of cells,
    oxygenation level, surviving fraction, etc.
    """

    header_file_extension = '.hed'
    data_file_extension = None
    allowed_suffix = tuple()

    def __init__(self, cube=None):
        if cube is not None:  # copying constructor
            self.header_set = cube.header_set
            self.version = cube.version
            self.modality = cube.modality
            self.created_by = cube.created_by
            self.creation_info = cube.creation_info
            self.primary_view = cube.primary_view
            self.data_type = cube.data_type
            self.num_bytes = cube.num_bytes
            self.byte_order = cube.byte_order
            self.patient_name = cube.patient_name
            self.patient_id = cube.patient_id
            self.slice_dimension = cube.slice_dimension
            self.pixel_size = cube.pixel_size
            self.slice_distance = cube.slice_distance
            self.slice_thickness = cube.slice_thickness
            self.slice_number = cube.slice_number
            self.xoffset = cube.xoffset  # self.xoffset are in mm, synced with DICOM contours
            self.dimx = cube.dimx
            self.yoffset = cube.yoffset  # self.yoffset are in mm, synced with DICOM contours
            self.dimy = cube.dimy
            self.zoffset = cube.zoffset  # self.zoffset are in mm, synced with DICOM contours
            self.dimz = cube.dimz
            self.z_table = cube.z_table
            self.slice_pos = cube.slice_pos
            self.basename = cube.basename
            self._set_format_str()
            self._set_number_of_bytes()

            # unique for whole structure set
            self._dicom_study_instance_uid = cube._dicom_study_instance_uid
            self._ct_dicom_series_instance_uid = cube._ct_dicom_series_instance_uid

            # unique for each CT slice
            self._ct_sop_instance_uid = cube._ct_sop_instance_uid

            self.cube = np.zeros((self.dimz, self.dimy, self.dimx), dtype=cube.pydata_type)

        else:
            import getpass
            from pytrip import __version__ as _ptversion

            self.header_set = False
            self.version = "2.0"
            self.modality = "CT"
            self.created_by = getpass.getuser()
            self.creation_info = "Created with PyTRiP98 {:s}".format(_ptversion)
            self.primary_view = "transversal"  # e.g. transversal
            self.data_type = ""
            self.num_bytes = ""
            self.byte_order = "vms"  # aix or vms
            self.patient_name = ""
            self.patient_id = datetime.datetime.today().strftime('%Y%m%d-%H%M%S')  # create a new patient ID if absent
            self.slice_dimension = ""  # eg. 256 meaning 256x256 pixels.
            self.pixel_size = ""  # size in [mm]
            self.slice_distance = ""  # distance between slices in [mm]
            self.slice_thickness = ""  # thickness of slice (usually equal to slice_distance) in [mm]
            self.slice_number = ""  # number of slices in file.
            self.xoffset = 0.0
            self.dimx = ""  # number of pixels along x (e.g. 256)
            self.yoffset = 0.0
            self.dimy = ""
            self.zoffset = 0.0
            self.dimz = ""
            self.slice_pos = []
            self.basename = ""

            # UIDs unique for whole structure set
            # generation of UID is done here in init, the reason why we are not generating them in create_dicom
            # method is that subsequent calls to write method shouldn't changed UIDs
            self._dicom_study_instance_uid = uid.generate_uid(prefix=None)
            self._ct_dicom_series_instance_uid = uid.generate_uid(prefix=None)

            # unique for each CT slice
            self._ct_sop_instance_uid = uid.generate_uid(prefix=None)

            self.z_table = False  # positions are stored in self.slice_pos (list of slice#,pos(mm),thickness(mm),tilt)

    def __add__(self, other):
        """ Overload + operator
        """
        c = type(self)(self)
        if Cube in other.__class__.__bases__:
            c.cube = other.cube + self.cube
        else:
            c.cube = self.cube + float(other)
        return c

    def __sub__(self, other):
        """ Overload - operator
        """
        c = type(self)(self)
        if Cube in other.__class__.__bases__:
            c.cube = self.cube - other.cube
        else:
            c.cube = self.cube - float(other)
        return c

    def __mul__(self, other):
        """ Overload * operator
        """
        c = type(self)(self)
        if Cube in other.__class__.__bases__:
            c.cube = other.cube * self.cube
        else:
            t = type(c.cube[0, 0, 0])
            c.cube = np.array(self.cube * float(other), dtype=t)
        return c

    def __div__(self, other):
        """ Overload / operator
        """
        c = type(self)(self)
        if Cube in other.__class__.__bases__:
            c.cube = self.cube / other.cube
        else:
            t = type(c.cube[0, 0, 0])
            c.cube = np.array(self.cube / float(other), dtype=t)
        c.cube[np.isnan(c.cube)] = 0  # fix division by zero NaNs
        return c

    __truediv__ = __div__

    # TODO __floordiv__ should also be handled

    def is_compatible(self, other):
        """ Check if this Cube object is compatible in size and dimensions with 'other' cube.

        A cube object can be a CtxCube, DosCube, LETCube or similar object.
        Unlike check_compatibility(), this function compares itself to the other cube.

        :param Cube other: The other Cube object which will be checked compatibility with.
        :returns: True if compatible.
        """
        return self.check_compatibility(self, other)

    @staticmethod
    def check_compatibility(a, b):
        """
        Simple comparison of cubes. if X,Y,Z dims are the same, and
        voxel sizes as well, then they are compatible. (Duck typed)

        See also the function is_compatible().

        :params Cube a: the first cube to be compared with the second (b).
        :params Cube b: the second cube to be compared with the first (a).

        """
        eps = 1e-5

        if a.dimx != b.dimx:
            return False
        elif a.dimy != b.dimy:
            return False
        elif a.dimz != b.dimz:
            return False
        elif (a.pixel_size - b.pixel_size) > eps:
            return False
        elif a.slice_distance != b.slice_distance:
            return False
        else:
            return True

    def indices_to_pos(self, indices):
        """ Translate index number of a voxel to real position in [mm], including any offsets.

        The z position is always following the slice positions.

        :params [int] indices: tuple or list of integer indices (i,j,k) or [i,j,k]
        :returns: list of positions, including offsets, as a list of floats [x,y,z]
        """
        pos = [(indices[0] + 0.5) * self.pixel_size + self.xoffset,
               (indices[1] + 0.5) * self.pixel_size + self.yoffset,
               self.slice_pos[indices[2]]]
        logger.debug("Map [i,j,k] {:d} {:d} {:d} to [x,y,z] {:.2f} {:.2f} {:.2f}".format(indices[0],
                                                                                         indices[1],
                                                                                         indices[2],
                                                                                         pos[0],
                                                                                         pos[1],
                                                                                         pos[2]))
        return pos

    def slice_to_z(self, slice_number):
        """ Return z-position in [mm] of slice number (starting at 1).

        :params int slice_number: slice number, starting at 1 and no bound check done here.
        :returns: position of slice in [mm] including offset
        """
        # note that self.slice_pos contains an array of positions including any zoffset.
        return self.slice_pos[slice_number - 1]

    def create_cube_from_equation(self, equation, center, limits, radial=True):
        """ Create Cube from a given equation.

        This function is currently out of order.

        """
        # TODO why eq not being used ?
        # eq = util.evaluator(equation)
        # TODO why data not being used ?
        # data = np.array(np.zeros((self.dimz, self.dimy, self.dimx)))
        x = np.linspace(0.5, self.dimx - 0.5, self.dimx) * self.pixel_size - center[0]
        y = np.linspace(self.dimx - 0.5, 0.5, self.dimx) * self.pixel_size - center[1]
        xv, yv = np.meshgrid(x, y)

    def mask_by_voi_all(self, voi, preset=0, data_type=np.int16):
        """ Attaches/overwrites Cube.data based on a given Voi.

        Voxels within the structure are filled it with 'preset' value.
        Voxels outside the contour will be filled with Zeros.

        :param Voi voi: the volume of interest
        :param int preset: value to be assigned to the voxels within the contour.
        :param data_type: numpy data type, default is np.int16

        TODO: this needs some faster implementation.
        """
        data = np.array(np.zeros((self.dimz, self.dimy, self.dimx)), dtype=data_type)
        if preset != 0:
            for i_z in range(self.dimz):
                for i_y in range(self.dimy):
                    # For a line along y, figure out how many contour intersections there are,
                    # then check how many intersections there are with x < than current point.
                    # If the number is odd, then the point is inside the VOI.
                    # If the number is even, then the point is outisde the VOI.
                    # This algorithm also works with multiple disconnected contours.
                    intersection = voi.get_row_intersections(self.indices_to_pos([0, i_y, i_z]))
                    if intersection is None:
                        break
                    if len(intersection) > 0:
                        k = 0
                        for i_x in range(self.dimx):
                            # count the number of intersections k along y, where intersection_x < current x position
                            if self.indices_to_pos([i_x, 0, 0])[0] > intersection[k]:
                                k += 1
                                if k >= len(intersection):
                                    break
                            if k % 2 == 1:  # voxel is inside structure, if odd number of intersections.
                                data[i_z][i_y][i_x] = preset
        self.cube = data

    def create_empty_cube(self, value, dimx, dimy, dimz, pixel_size, slice_distance, slice_offset=0.0):
        """ Creates an empty Cube object.

        Values are stored as 2-byte integers.

        :param int16 value: integer value which will be assigned to all voxels.
        :param int dimx: number of voxels along x
        :param int dimy: number of voxels along y
        :param int dimz: number of voxels along z
        :param float pixel_size: size of each pixel (x == y) in [mm]
        :param float slice_distance: the distance between two slices (z) in [mm]
        :param float slice_offset: start position of the first slice in [mm] (default 0.0 mm)
        """
        self.dimx = dimx
        self.dimy = dimy
        self.dimz = dimz
        self.slice_number = dimz
        self.pixel_size = pixel_size
        self.slice_distance = slice_distance
        self.slice_thickness = slice_distance  # use distance for thickness as default
        self.cube = np.ones((dimz, dimy, dimx), dtype=np.int16) * value
        self.slice_dimension = dimx
        self.num_bytes = 2
        self.data_type = "integer"
        self.pydata_type = np.int16
        self.slice_pos = [slice_distance * i + slice_offset for i in range(dimz)]
        self.header_set = True
        self.patient_id = ''
        # UIDs unique for whole structure set
        # generation of UID is done here in init, the reason why we are not generating them in create_dicom
        # method is that subsequent calls to write method shouldn't changed UIDs
        self._dicom_study_instance_uid = uid.generate_uid(prefix=None)
        self._ct_dicom_series_instance_uid = uid.generate_uid(prefix=None)
        # unique for each CT slice
        self._ct_sop_instance_uid = uid.generate_uid(prefix=None)

    def mask_by_voi(self, voi, value):
        """ Overwrites the Cube voxels within the given Voi with 'value'.

        Voxels within the structure are filled it with 'value'.
        Voxels outside the contour are not touched.

        :param Voi voi: the volume of interest
        :param value=0: value to be assigned to the voxels within the contour.
        """
        for i_z in range(self.dimz):
            for i_y in range(self.dimy):
                intersection = voi.get_row_intersections(self.indices_to_pos([0, i_y, i_z]))
                if intersection is None:
                    break
                if len(intersection) > 0:
                    k = 0
                    for i_x in range(self.dimx):
                        if self.indices_to_pos([i_x, 0, 0])[0] > intersection[k]:
                            k += 1
                            if k >= (len(intersection)):
                                break
                        if k % 2 == 1:  # voxel is inside structure, if odd number of intersections.
                            self.cube[i_z][i_y][i_x] = value

    def mask_by_voi_add(self, voi, value=0):
        """ Add 'value' to all voxels within the given Voi

        'value' is added to each voxel value within the given volume of interest.
        Voxels outside the volume of interest are not touched.

        :param Voi voi: the volume of interest
        :param value=0: value to be added to the voxel values within the contour.
        """
        for i_z in range(self.dimz):
            for i_y in range(self.dimy):
                intersection = voi.get_row_intersections(self.indices_to_pos([0, i_y, i_z]))
                if intersection is None:
                    break
                if len(intersection) > 0:
                    k = 0
                    for i_x in range(self.dimx):
                        if self.indices_to_pos([i_x, 0, 0])[0] > intersection[k]:
                            k += 1
                            if k >= (len(intersection)):
                                break
                        if k % 2 == 1:  # voxel is inside structure, if odd number of intersections.
                            self.cube[i_z][i_y][i_x] += value

    def merge(self, cube):
        self.cube = np.maximum(self.cube, cube.cube)

    def merge_zero(self, cube):
        self.cube[self.cube == 0] = cube.cube[self.cube == 0]

    # ######################  READING TRIP98 FILES #######################################

    @classmethod
    def header_file_name(cls, path_name):
        return TRiP98FilePath(path_name, cls).header

    @classmethod
    def data_file_name(cls, path_name):
        return TRiP98FilePath(path_name, cls).datafile

    def read(self, path):
        """
        Reads both TRiP98 data and its associated header into the Cube object.

        Cube can be read providing a filename stripped of extension, i.e:
        >>> import pytrip as pt
        >>> c1 = pt.CtxCube()
        >>> c1.read("tests/res/TST003/tst003000")

        We can also read header file and data path which do not share common basename:
        >>> c2 = pt.CtxCube()
        >>> c2.read(("tests/res/TST003/tst003012.hed", "tests/res/TST003/tst003000.ctx.gz"))

        :param path: string or sequence of strings (length 2)
        :return:
        """

        # let us check if path is a string in a way python 2 and 3 will like it
        # based on https://stackoverflow.com/questions/4843173/how-to-check-if-type-of-a-variable-is-string
        running_python2 = sys.version_info.major == 2
        path_string = isinstance(path, (str, bytes) if not running_python2 else basestring)  # NOQA: F821

        # single argument of string type, i.e. filename without extension
        if path_string:
            self.basename = os.path.basename(TRiP98FilePath(path, self).basename)

            path_locator = TRiP98FileLocator(path, self)

            header_path = path_locator.header
            datafile_path = path_locator.datafile

            if not datafile_path or not header_path:
                raise FileNotFound("Loading {:s} failed, file not found".format(path))

        # tuple with path to header and datafile
        elif len(path) == 2:
            header_path, datafile_path = path

            # security checks for header file
            # first check - validity of the path
            if not TRiP98FilePath(header_path, self).is_valid_header_path():
                logger.warning("Loading {:s} which doesn't look like valid header path".format(header_path))

            # second check - if file exists
            if not os.path.exists(header_path):
                header_path_locator = TRiP98FileLocator(header_path, self)
                if header_path_locator.header is not None:
                    logger.warning("Did you meant to load {:s}, instead of {:s} ?".format(header_path_locator.header,
                                                                                          header_path))
                raise FileNotFound("Loading {:s} failed, file not found".format(header_path))

            # security checks for datafile path
            # first check - validity of the path
            if not TRiP98FilePath(datafile_path, self).is_valid_datafile_path():
                logger.warning("Loading {:s} which doesn't look like valid datafile path".format(datafile_path))

            # second check - if file exists
            if not os.path.exists(datafile_path):
                datafile_path_locator = TRiP98FileLocator(datafile_path, self)
                if datafile_path_locator.datafile is not None:
                    logger.warning(
                        "Did you meant to load {:s}, instead of {:s} ?".format(datafile_path_locator.datafile,
                                                                               datafile_path))
                raise FileNotFound("Loading {:s} failed, file not found".format(datafile_path))

            self.basename = ""  # TODO user may provide two completely different filenames for header and datafile
            # i.e. read( ("1.hed", "2.dos"), what about basename then ?

        else:
            raise Exception("More than two arguments provided as path variable to Cube.read method")

        # finally read files
        self._read_trip_header_file(header_path=header_path)
        self._read_trip_data_file(datafile_path=datafile_path, header_path=header_path)

    def _read_trip_header_file(self, header_path):  # TODO: could be made private? #126
        """ Reads a header file, accepts also if file is .gz compressed.
        First the un-zipped files will be attempted to read.
        Should these not exist, then the .gz are attempted.

        However, if the .hed.gz file was explicitly stated,
        then this file will also be loaded, even if a .hed is available.
        """

        # sanity check
        if header_path is not None:
            logger.info("Reading header file" + header_path)
        else:
            raise IOError("Could not find file " + header_path)

        # load plain of gzipped file
        if header_path.endswith(".gz"):
            import gzip
            fp = gzip.open(header_path, "rt")
        else:
            fp = open(header_path, "rt")
        content = fp.read()
        fp.close()

        # fill self with data
        self._parse_trip_header(content)
        self._set_format_str()
        logger.debug("Format string:" + self.format_str)

    def _read_trip_data_file(self, datafile_path, header_path,
                             multiply_by_2=False):  # TODO: could be made private? #126
        """Read TRiP98 formatted data.

        If header file was not previously loaded, it will be attempted first.

        Due to an issue in VIRTUOS, sometimes DosCube data have been reduced with a factor of 2.
        Setting multiply_by_2 to True, will restore the true values, in this case.

        :param datafile_path: Path to TRiP formatted data.
        :param multiply_by_2: The data read will automatically be multiplied with a factor of 2.
        """

        # fill header data if self.header is empty
        if not self.header_set:
            self._read_trip_header_file(header_path)

        # raise exception if reading header failed
        if not self.header_set:
            raise InputError("Header file not loaded")

        # preparation
        data_dtype = np.dtype(self.format_str)
        data_count = self.dimx * self.dimy * self.dimz

        # load data from data file (gzipped or not)
        logger.info("Opening file: " + datafile_path)
        if datafile_path.endswith('.gz'):
            import gzip
            with gzip.open(datafile_path, "rb") as f:
                s = f.read(data_dtype.itemsize * data_count)
                tmpcube = np.frombuffer(s, dtype=data_dtype, count=data_count)
                # frombuffer returns read-only array, so we need to make it writable
                cube = np.require(tmpcube, dtype=data_dtype, requirements=['W', 'O'])
        else:
            cube = np.fromfile(datafile_path, dtype=data_dtype)

        if self.byte_order == "aix":
            logger.info("AIX big-endian data.")
            # byteswapping is not needed anymore, handled by "<" ">" in dtype

        # sanity check
        logger.info("Cube data points : {:d}".format(len(cube)))
        if len(cube) != self.dimx * self.dimy * self.dimz:
            logger.error("Header size and cube size dont match.")
            logger.error("Cube data points : {:d}".format(len(cube)))
            logger.error("Header says      : {:d} = {:d} * {:d} * {:d}".format(
                self.dimx * self.dimy * self.dimz,
                self.dimx,
                self.dimy,
                self.dimz))
            raise IOError("Header data and dose cube size are not consistent.")

        cube = np.reshape(cube, (self.dimz, self.dimy, self.dimx))
        if multiply_by_2:
            logger.warning("Cube was previously rescaled to 50%. Now multiplying with 2.")
            cube *= 2
        self.cube = cube

    def _parse_trip_header(self, content):
        """ Parses content which was read from a trip header.
        """
        i = 0
        self.header_set = True
        content = content.split('\n')
        self.z_table = False
        while i < len(content):
            if re.match("version", content[i]):
                self.version = content[i].split()[1]
            if re.match("modality", content[i]):
                self.modality = content[i].split()[1]
            if re.match("created_by", content[i]):
                self.created_by = content[i].replace("created_by ", "", 1)
                self.created_by = self.created_by.rstrip()
            if re.match("creation_info", content[i]):
                self.creation_info = content[i].replace("creation_info ", "", 1)
                self.creation_info = self.creation_info.rstrip()
            if re.match("primary_view", content[i]):
                self.primary_view = content[i].split()[1]
            if re.match("data_type", content[i]):
                self.data_type = content[i].split()[1]
            if re.match("num_bytes", content[i]):
                self.num_bytes = int(content[i].split()[1])
            if re.match("byte_order", content[i]):
                self.byte_order = content[i].split()[1]
            if re.match("patient_name", content[i]):
                self.patient_name = content[i].split()[1]
            if re.match("slice_dimension", content[i]):
                self.slice_dimension = int(content[i].split()[1])
            if re.match("pixel_size", content[i]):
                self.pixel_size = float(content[i].split()[1])
            if re.match("slice_distance", content[i]):
                self.slice_distance = float(content[i].split()[1])
                self.slice_thickness = float(content[i].split()[1])  # TRiP format only. See #342
            if re.match("slice_number", content[i]):
                self.slice_number = int(content[i].split()[1])
            if re.match("xoffset", content[i]):
                self.xoffset = int(content[i].split()[1])
            if re.match("yoffset", content[i]):
                self.yoffset = int(content[i].split()[1])
            if re.match("zoffset", content[i]):
                self.zoffset = int(content[i].split()[1])
            if re.match("dimx", content[i]):
                self.dimx = int(content[i].split()[1])
            if re.match("dimy", content[i]):
                self.dimy = int(content[i].split()[1])
            if re.match("dimz", content[i]):
                self.dimz = int(content[i].split()[1])
            if re.match("slice_no", content[i]):
                self.slice_pos = [float(j) for j in range(self.slice_number)]
                self.z_table = True
                i += 1
                for j in range(self.slice_number):
                    self.slice_pos[j] = float(content[i].split()[1])
                    i += 1
            i += 1

        # zoffset from TRiP contains the integer amount of slice thicknesses as offset.
        # Here we convert to an actual offset in mm, which is stored in self
        self.xoffset *= self.pixel_size
        self.yoffset *= self.pixel_size
        self.zoffset *= self.slice_distance

        logger.debug("TRiP loaded offsets: {:f} {:f} {:f}".format(self.xoffset, self.yoffset, self.zoffset))

        # generate slice position tables, if absent in header file
        # Note:
        # - ztable in .hed is _without_ offset
        # - self.slice_pos however holds values _including_ offset.
        if not self.z_table:
            self.slice_pos = [self.zoffset + _i * self.slice_distance for _i in range(self.slice_number)]
        self._set_format_str()

    def _set_format_str(self):
        """Set format string according to byte_order.
        """
        if self.byte_order == "vms":
            self.format_str = "<"
        elif self.byte_order == "aix":
            self.format_str = ">"
        self._set_number_of_bytes()

    def _set_number_of_bytes(self):
        """Set format_str and pydata_type according to num_bytes and data_type
        """
        if self.data_type == "integer":
            if self.num_bytes == 1:
                self.format_str += "b"
                self.pydata_type = np.int8
            if self.num_bytes == 2:
                self.format_str += "h"
                self.pydata_type = np.int16
            if self.num_bytes == 4:
                self.format_str += "i"
                self.pydata_type = np.int32
        elif self.data_type in ["float", "double"]:
            if self.num_bytes == 4:
                self.format_str += "f"
                self.pydata_type = np.float32
            if self.num_bytes == 8:
                self.format_str += "d"
                self.pydata_type = np.double
        else:
            print("Format:", self.byte_order, self.data_type, self.num_bytes)
            raise IOError("Unsupported format.")
        logger.debug("self.format_str: '{}'".format(self.format_str))

    # ######################  WRITING TRIP98 FILES #######################################

    def write(self, path):
        """Write the Cube and its header to a file with the filename 'path'.

        :param str path: path to header file, data file or basename (without extension)
        :returns: tuple header_path, datafile_path: paths to header file and datafiles where data was saved
        (may be different from input path if user provided a partial basename)
        """

        running_python2 = sys.version_info.major == 2
        path_string = isinstance(path, (str, bytes) if not running_python2 else basestring)  # NOQA: F821

        if path_string:
            header_path = self.header_file_name(path)
            datafile_path = self.data_file_name(path)

        elif len(path) == 2:
            header_path, datafile_path = path

            # security checks for header file
            # first check - validity of the path
            if not TRiP98FilePath(header_path, self).is_valid_header_path():
                logger.warning("Loading {:s} which doesn't look like valid header path".format(header_path))

            # security checks for datafile path
            # first check - validity of the path
            if not TRiP98FilePath(datafile_path, self).is_valid_datafile_path():
                logger.warning("Loading {:s} which doesn't look like valid datafile path".format(datafile_path))
        else:
            raise Exception("More than two arguments provided as path variable to Cube.read method")

        # finally write files
        self._write_trip_header(header_path)
        self._write_trip_data(datafile_path)

        return header_path, datafile_path

    def _write_trip_header(self, path):
        """ Write a TRiP98 formatted header file, based on the available meta data.

        :param path: fully qualified path, including file extension (.hed)
        """
        from distutils.version import LooseVersion
        output_str = "version " + self.version + "\n"
        output_str += "modality " + self.modality + "\n"
        # include created_by and creation_info only for files newer than 1.4
        if LooseVersion(self.version) >= LooseVersion("1.4"):
            output_str += "created_by {:s}\n".format(self.created_by)
            output_str += "creation_info {:s}\n".format(self.creation_info)
        output_str += "primary_view " + self.primary_view + "\n"
        output_str += "data_type " + self.data_type + "\n"
        output_str += "num_bytes " + str(self.num_bytes) + "\n"
        output_str += "byte_order " + self.byte_order + "\n"
        if self.patient_name == "":
            self.patient_name = "Anonymous"
        # patient_name in .hed must be equal to the base filename without extension, else TRiP98 wont import VDX
        _fname = os.path.basename(path)
        _pname = os.path.splitext(_fname)[0]
        output_str += "patient_name {:s}\n".format(_pname)
        output_str += "slice_dimension {:d}\n".format(self.slice_dimension)
        output_str += "pixel_size {:.7f}\n".format(self.pixel_size)
        output_str += "slice_distance {:.7f}\n".format(self.slice_distance)
        output_str += "slice_number " + str(self.slice_number) + "\n"
        output_str += "xoffset {:d}\n".format(int(round(self.xoffset / self.pixel_size)))
        output_str += "dimx {:d}\n".format(self.dimx)
        output_str += "yoffset {:d}\n".format(int(round(self.yoffset / self.pixel_size)))
        output_str += "dimy {:d}\n".format(self.dimy)

        # zoffset in Voxelplan .hed seems to be broken, and should not be used if not = 0
        # to apply zoffset, z_table should be used instead.
        # This means, self.zoffset should not be used anywhere.
        output_str += "zoffset 0\n"
        output_str += "dimz " + str(self.dimz) + "\n"
        if self.z_table:
            output_str += "z_table yes\n"
            output_str += "slice_no  position  thickness  gantry_tilt\n"
            for i, item in enumerate(self.slice_pos):
                output_str += "  {:<3d}{:14.4f}{:13.4f}{:14.4f}\n".format(i + 1, item,
                                                                          self.slice_thickness, 0)  # 0 gantry tilt
        else:
            output_str += "z_table no\n"

        with open(path, "w+") as f:
            f.write(output_str)

    def _write_trip_data(self, path):
        """ Writes the binary data cube in TRiP98 format to a file.

        Type is specified by self.pydata_type and self.byte_order attributes.

        :param str path: Full path including file extension.
        """
        cube = np.array(self.cube, dtype=self.pydata_type)
        if self.byte_order == "aix":
            cube = cube.byteswap()
        cube.tofile(path)

    # ######################  READING DICOM FILES #######################################

    def _set_z_table_from_dicom(self, dcm):
        """ Creates the slice position lookup table based on a given Dicom object.
        The table is attached to self.

        :param DICOM dcm: DICOM object provided by pydicom.
        """
        # TODO: can we rely on that this will always be sorted?
        # if yes, then all references to whether this is sorted or not can be removed hereafter
        # (see also pytripgui) /NBassler
        self.slice_pos = []
        for i, dcm_image in enumerate(dcm["images"]):
            self.slice_pos.append(float(dcm_image.ImagePositionPatient[2]))

    def _set_header_from_dicom(self, dcm):
        """ Creates the header metadata for this Cube class, based on a given Dicom object.

        :param DICOM dcm: Dicom object which will be used for generating the header data.
        """
        if not _dicom_loaded:
            raise ModuleNotLoadedError("Dicom")
        ds = dcm["images"][0]
        self.version = "1.4"
        self.created_by = "pytrip"
        self.creation_info = "Created by PyTRiP98;"
        self.primary_view = "transversal"
        self.set_data_type(type(ds.pixel_array[0][0]))
        self.patient_name = ds.PatientName
        self.basename = ds.PatientID.replace(" ", "_")
        self.slice_dimension = int(ds.Rows)  # should be changed ?
        self.pixel_size = float(ds.PixelSpacing[0])  # (0028, 0030) Pixel Spacing (DS)
        self.slice_thickness = ds.SliceThickness  # (0018, 0050) Slice Thickness (DS)
        # slice_distance != SliceThickness. One may have overlapping slices. See #342
        self.slice_number = len(dcm["images"])
        self.xoffset = float(ds.ImagePositionPatient[0])
        self.dimx = int(ds.Rows)  # (0028, 0010) Rows (US)
        self.yoffset = float(ds.ImagePositionPatient[1])
        self.dimy = int(ds.Columns)  # (0028, 0011) Columns (US)
        self.zoffset = float(ds.ImagePositionPatient[2])  # note that zoffset should not be used.
        self.dimz = len(dcm["images"])
        self._set_z_table_from_dicom(dcm)
        self.z_table = True

        # Fix for bug #342
        # TODO: slice_distance should probably be a list of distances,
        # but for now we will just use the distance between the first two slices.
        if len(self.slice_pos) > 1:  # _set_z_table_from_dicom() must be called before
            self.slice_distance = abs(self.slice_pos[1] - self.slice_pos[0])
            logger.debug("Slice distance set to {:.2f}".format(self.slice_distance))
        else:
            logger.warning("Only a single slice found. Setting slice_distance to slice_thickness.")
            self.slice_distance = self.slice_thickness

        if self.slice_thickness > self.slice_distance:
            # TODO: this is probably valid dicom format, however let's print a warning for now
            # as it may indicate some problem with the input dicom, as it is rather unusual.
            logger.warning("Overlapping slices found: slice thickness is larger than the slice distance.")

        self.set_byteorder()
        self.data_type = "integer"
        self.num_bytes = 2
        self._set_format_str()
        self.header_set = True

        # unique for whole structure set
        self._dicom_study_instance_uid = ds.StudyInstanceUID
        self._ct_dicom_series_instance_uid = ds.SeriesInstanceUID

    def set_byteorder(self, endian=None):
        """Set/change the byte order of the data to be written to disk.

        Available options are:
        - 'little' vms, Intel style little-endian byte order.
        - 'big' aix, Motorola style big-endian byte order.
        - if unspecified, the native system dependent endianess is used.

        :param str endian: optional string containing the endianess.
        """
        if endian is None:
            endian = sys.byteorder
        if endian == 'little':
            self.byte_order = "vms"
        elif endian == 'big':
            self.byte_order = "aix"
        else:
            raise ValueError("set_byteorder error: unknown endian " + str(endian))

    def set_data_type(self, type):
        """ Sets the data type for the TRiP98 header files.

        :param numpy.type type: numpy type, e.g. np.uint16
        """
        if type is np.int8 or type is np.uint8:
            self.data_type = "integer"
            self.num_bytes = 1
        elif type is np.int16 or type is np.uint16:
            self.data_type = "integer"
            self.num_bytes = 2
        elif type is np.int32 or type is np.uint32:
            self.data_type = "integer"
            self.num_bytes = 4
        elif type is np.float:
            self.data_type = "float"
            self.num_bytes = 4
        elif type is np.double:
            self.data_type = "double"
            self.num_bytes = 8

    # ######################  WRITING DICOM FILES #######################################

    def create_dicom_base(self):
        if _dicom_loaded is False:
            raise ModuleNotLoadedError("Dicom")
        if self.header_set is False:
            raise InputError("Header not loaded")

        # TODO tags + code datatypes are described here:
        # https://www.dabsoft.ch/dicom/6/6/#(0020,0012)
        # datatype codes are described here:
        # ftp://dicom.nema.org/medical/DICOM/2013/output/chtml/part05/sect_6.2.html

        meta = Dataset()
        meta.MediaStorageSOPClassUID = '1.2.840.10008.5.1.4.1.1.2'  # CT Image Storage
        # Media Storage SOP Instance UID tag 0x0002,0x0003 (type UI - Unique Identifier)
        meta.MediaStorageSOPInstanceUID = self._ct_sop_instance_uid
        meta.ImplementationClassUID = "1.2.3.4"
        meta.TransferSyntaxUID = uid.ImplicitVRLittleEndian  # Implicit VR Little Endian - Default Transfer Syntax
        ds = FileDataset("file", {}, file_meta=meta, preamble=b"\0" * 128)
        ds.PatientName = self.patient_name
        if self.patient_id in (None, ''):
            ds.PatientID = datetime.datetime.today().strftime('%Y%m%d-%H%M%S')
        else:
            ds.PatientID = self.patient_id  # Patient ID tag 0x0010,0x0020 (type LO - Long String)
        ds.PatientSex = ''  # Patient's Sex tag 0x0010,0x0040 (type CS - Code String)
        #                      Enumerated Values: M = male F = female O = other.
        ds.PatientBirthDate = '19010101'
        ds.SpecificCharacterSet = 'ISO_IR 100'
        ds.AccessionNumber = ''
        ds.is_little_endian = True
        ds.is_implicit_VR = True
        ds.SOPClassUID = '1.2.3'  # !!!!!!!!
        # SOP Instance UID tag 0x0008,0x0018 (type UI - Unique Identifier)
        ds.SOPInstanceUID = self._ct_sop_instance_uid

        # Study Instance UID tag 0x0020,0x000D (type UI - Unique Identifier)
        # self._dicom_study_instance_uid may be either set in __init__ when creating new object
        #   or set when import a DICOM file
        #   Study Instance UID for structures is the same as Study Instance UID for CTs
        ds.StudyInstanceUID = self._dicom_study_instance_uid

        # Series Instance UID tag 0x0020,0x000E (type UI - Unique Identifier)
        # self._ct_dicom_series_instance_uid may be either set in __init__ when creating new object
        #   or set when import a DICOM file
        #   Series Instance UID for structures might be different than Series Instance UID for CTs
        ds.SeriesInstanceUID = self._ct_dicom_series_instance_uid

        # Study Instance UID tag 0x0020,0x000D (type UI - Unique Identifier)
        ds.FrameofReferenceUID = '1.2.3'  # !!!!!!!!!
        ds.StudyDate = datetime.datetime.today().strftime('%Y%m%d')
        ds.StudyTime = datetime.datetime.today().strftime('%H%M%S')
        ds.PhotometricInterpretation = 'MONOCHROME2'
        ds.SamplesPerPixel = 1
        ds.ImageOrientationPatient = ['1', '0', '0', '0', '1', '0']
        ds.Rows = self.dimx
        ds.Columns = self.dimy
        ds.SliceThickness = str(self.slice_distance)
        ds.PixelSpacing = [self.pixel_size, self.pixel_size]

        # Add eclipse friendly IDs
        ds.StudyID = '1'  # Study ID tag 0x0020,0x0010 (type SH - Short String)
        ds.ReferringPhysiciansName = 'py^trip'  # Referring Physician's Name tag 0x0008,0x0090 (type PN - Person Name)
        ds.PositionReferenceIndicator = ''  # Position Reference Indicator tag 0x0020,0x1040
        ds.SeriesNumber = '1'  # SeriesNumber tag 0x0020,0x0011 (type IS - Integer String)

        return ds



#
#    Copyright (C) 2010-2017 PyTRiP98 Developers.
#
#    This file is part of PyTRiP98.
#
#    PyTRiP98 is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyTRiP98 is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyTRiP98.  If not, see <http://www.gnu.org/licenses/>.
#
"""
The RCR model is based on the paper from Antonovic et al.
https://doi.org/10.1093/jrr/rru020
Parameters are set for C-12 ions only.
"""

import numpy as np
import logging

logger = logging.getLogger(__name__)


def rbe_rcr(dose_ion, let, alpha_x, beta_x, oxy=None):
    """
    Returns the RBE for a given dose/let cube.

    input parameters may be either numpy.array or scalars
    TODO: handle real cubes.

    :params dose_ion: ion physical dose in [Gy]
    :params let: LET in [keV/um]
    :params alpha_x: alpha for X-rays in [Gy^-1]
    :params beta_x: beta for X-rays in [Gy^-2]
    :params oxy: optional oxygenation cube in [mmHgO_2]
    """

    # Calculate sf_ion(D_ion, let, oxy)
    # from pytrip.models.extra import rbe_from_sf
    # rbe_from_sf(_sf, dose_ion, alpha_x, beta_x)
    logger.warning("rcr_rbe not implemented yet.")
    pass  # TODO: not implemented yet.


def sf_rcr(dose, let, oxy=None):
    """
    Function which returns surving fraction
    Equation (3) in https://doi.org/10.1093/jrr/rru020

    input parameters may be either numpy.array or scalars
    TODO: handle real cubes.

    :params dose: physical ion dose in [Gy]
    :params let: LET in keV/um
    :params oxy: optional oxygenation in [mmHgO_2]
    """

    a0 = 5.7
    a1 = 1.3
    b1 = 2.0
    c0 = 5.7
    c1 = 0.2
    ln = 423.0

    if oxy is None:
        ocube = 1.0
    else:
        ocube = oer_po2_rcr(let, oxy)
        # ocube = oer_rcr(let)  # TODO: this was old version, not sure how to include.

    a = (a0 * _f(let) + a1 * np.exp(-let / ln)) / ocube
    b = b1 * np.exp(-let / ln) / ocube
    c = (c0 * _f(let) + c1 * np.exp(-let / ln)) / ocube

    sf = np.exp(-dose * a) + dose * b * np.exp(-dose * c)
    return sf


def _f(let):
    """
    f function from Dasu paper, takes let-cube as parameter
    Equation (7) in https://doi.org/10.1093/jrr/rru020

    input parameters may be either numpy.array or scalars
    TODO: handle real cubes.

    :params let: LET in [keV/um]

    :returns: result of the f function
    """

    ld = 86.0
    result = (1 - np.exp(-let / ld) * (1 + let / ld)) * ld / let

    # map any zero LET areas to 0.0
    if np.isscalar(result):  # scalar
        if result == np.inf:
            result = 0.0
    else:
        result[result == np.inf] = 0.0  # numpy arrays

    return result


def oer_rcr(let):
    """
    ~O dose modifying factor.
    Equation (2) in https://doi.org/10.1093/jrr/rru020

    input parameters may be either numpy.array or scalars
    TODO: handle real cubes.

    :params let: LET in [keV/um]

    :returns: cube containing the oxygen enhancement ratio
    """

    omin = 1.10
    omax = 2.92
    lo = 114.0
    _o = omin + (omax - omin) * np.exp(-(let / lo) * (let / lo))
    return _o


def oer_po2_rcr(let, oxy):
    """
    ~O dose modifying factor, taking varying pO2 into account
    Equation (1) in https://doi.org/10.1093/jrr/rru020

    input parameters may be either numpy.array or scalars
    TODO: handle real cubes.

    :params let: LET in [keV/um]
    :params oxy: oxygenation in [mmHgO_2]

    :returns: cube containing the oxygen enhancement ratio
    """
    k = 2.5  # mmHg
    _o = oer_rcr(let) * (k + oer_rcr(let)) / (k + oer_rcr(let) * oxy)
    return _o



#
#    Copyright (C) 2010-2017 PyTRiP98 Developers.
#
#    This file is part of PyTRiP98.
#
#    PyTRiP98 is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyTRiP98 is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyTRiP98.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Collection of proton RBE models.

[1] A. Carabe, M. Moteabbed, N. Depauw, J. Schuemann, and H. Paganetti,
"Range uncertainty in proton therapy due to variable biological effectiveness,"
Phys. Med. Biol. 57(5), 1159-1172 (2012).
https://doi.org/10.1088/0031-9155/57/5/1159

[2] M. Wedenberg, B. Lind, and B. Haardemark,
"A model for the relative biological effectiveness of protons:
The tissue specific parameter alpha/beta of photons is a predictor for the sensitivity to LET changes,"
Acta Oncol. 52(3), 580-588 (2013).
http://dx.doi.org/10.3109/0284186X.2012.705892

[3]  A. L. McNamara, J. Schuemann, and H. Paganetti,
"A phenomenological relative biological effectiveness (RBE) model for proton therapy based on all published
in vitro cell survival data,"
Phys. Med. Biol. 60(21), 8399-8416
https://doi.org/10.1088/0031-9155/60/21/8399
"""

import logging
import numpy as np

logger = logging.getLogger(__name__)


def rbe_carabe(dose, let, abx):
    """
    Carabe proton RBE model

    input parameters may be either numpy.array or scalars
    TODO: handle Cube() class directly

    :params dose: physical proton dose in [Gy]
    :params let: LETd in [keV/um] (protons only)
    :params abx: alpha_x / beta_x [Gy]

    :returns: RBE for the given parameters

    :ref: https://doi.org/10.1088/0031-9155/57/5/1159
    """

    _labx = 2.686 * let / abx
    _apx = 0.843 + 0.154 * _labx
    _sbpx = 1.090 + 0.006 * _labx

    rbe = _rbe_apx(dose, _apx, _sbpx, abx)
    return rbe


def rbe_wedenberg(dose, let, abx):
    """
    Wedenberg proton RBE model

    input parameters may be either numpy.array or scalars
    TODO: handle Cube() class directly

    :params dose: physical proton dose in [Gy]
    :params let: LETd in [keV/um] (protons only)
    :params abx: alpha_x / beta_x [Gy]

    :returns: RBE for the given parameters

    :ref: http://dx.doi.org/10.3109/0284186X.2012.705892
    """

    _apx = 1.000 + 0.434 * let / abx
    _sbpx = 1.000

    rbe = _rbe_apx(dose, _apx, _sbpx, abx)
    return rbe


def rbe_mcnamara(dose, let, abx):
    """
    McNamara proton RBE model

    input parameters may be either numpy.array or scalars
    TODO: handle Cube() class directly

    :params dose: physical proton dose in [Gy]
    :params let: LETd in [keV/um] (protons only)
    :params abx: alpha_x / beta_x [Gy]

    :returns: RBE for the given parameters

    :ref: https://doi.org/10.1088/0031-9155/60/21/8399
    """

    _apx = 0.99064 + 0.35605 * let / abx
    _sbpx = 1.1012 - 0.0038703 * np.sqrt(abx) * let

    rbe = _rbe_apx(dose, _apx, _sbpx, abx)
    return rbe


def _rbe_apx(dose, apx, sbpx, abx, dzero=0.0):
    """
    :params dose: proton dose      [Gy]
    :params apx: alpha_p / alpha_x [dimensionless] RBE_max = ap/ax when (dose -> 0 Gy)
    :params sbpx: beta_p / beta_x  [dimensionless] RBE_min = sqrt(bp/bx) when (dose -> inf Gy)
    :params abx: alpha_x / beta_x  [Gy]
    :params dzero: what to return in case of dose is zero (which would cause division by zero)
    """

    _rbe = 1.0 / (2.0 * dose)
    if hasattr(_rbe, '__iter__'):
        _rbe[_rbe == np.inf] = dzero
    else:
        if _rbe == np.inf:
            return dzero
    _rbe *= (np.sqrt(abx * abx + 4 * abx * apx * dose + 4 * sbpx * sbpx * dose * dose) - abx)
    return _rbe



#
#    Copyright (C) 2010-2018 PyTRiP98 Developers.
#
#    This file is part of PyTRiP98.
#
#    PyTRiP98 is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyTRiP98 is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyTRiP98.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Object holding all data needed for a specific projectile/rifi configuration.
"""


class KernelModel(object):
    """
    Kernel data, holding all necessary data for a certain projectile
    """
    projectile = None
    rifi_name = ""  # name of ripple filter, e.g. "3 mm Rifi", for documentation only
    rifi_thickness = 0.0  # thickness of RiFi in mm
    ddd_path = ""  # "/local/trip/data/ddd/*"
    spc_path = ""  # "/local/trip/data/spc/*"
    sis_path = ""  # if none, it will be autogenerated
    comment = ""  # arbitrary text
    name = ""  # name for identifying this kernel, ie.e "C12_3mmRiFi"

    def __init__(self, name=""):
        """
        """
        self.name = name

    def __str__(self):
        """
        String out handler
        """
        return self._print()

    def _print(self):
        """
        Pretty print all attributes.
        """
        out = "\n"
        out += "   Kernel '{:s}'\n".format(self.name)
        out += "----------------------------------------------------------------------------\n"
        out += "|  RiFi Name                         : '{:s}'\n".format(str(self.rifi_name))
        out += "|  RiFi Thickness [mm]               : {:.2f}\n".format(self.rifi_thickness)
        out += "|"
        out += "|  DDD Path                          : '{:s}'\n".format(str(self.ddd_path))
        out += "|  SPC Path                          : '{:s}'\n".format(str(self.spc_path))
        out += "|  SIS Path                          : '{:s}'\n".format(str(self.sis_path))
        out += "|"
        out += "|  Comment                           : '\n{:s}'\n".format(str(self.comment))
        out += "----------------------------------------------------------------------------\n"
        if self.projectile:
            out += self.projectile._print()  # TODO: maybe _print should not be marked private.
        else:
            out += "| No projectile set.\n"
        return out



#
#    Copyright (C) 2010-2017 PyTRiP98 Developers.
#
#    This file is part of PyTRiP98.
#
#    PyTRiP98 is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyTRiP98 is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyTRiP98.  If not, see <http://www.gnu.org/licenses/>.
#
"""
This module provides the volume histogram class.
"""
import logging
import warnings
import numpy as np


class VolHist:
    """
    Volume histogram class
    """

    def __init__(self, cube, voi=None, target_dose=None):
        """
        :params Cube cube: either LETCube, DosCube or similar object.
        :params Voi voi: a single voi (vdx.vois[i])
        :target_dose: set target_dose in [Gy]. Any target_dose in cube.target_dose will be ignored, if set.
        """
        self.target_dose = target_dose
        self.cube_basename = cube.basename  # basename of the cube used for histogram
        self.x_is_relative = False  # relative or absolute x units

        self.name = "(none)"
        if voi:
            self.name = voi.name  # name of the VOI

        logging.info("Processing ROI '{:s}' for '{}'...".format(self.name, self.cube_basename))
        self.x, self.y = self.volume_histogram(cube.cube, voi)  # x,y data

        if not self.x.any() or not self.y.any():
            self.xlabel = "(no data)"
            self.ylabel = "(no data)"
            return

        self.ylabel = "Volume [%]"  # units on y-axis

        if cube.type == 'DOS':  # DOS Cube
            if not target_dose:
                target_dose = cube.target_dose  # optional target dose scaling factor
            # DOS cube data are stored in %%.
            if target_dose:
                _tdose = 0.001 * target_dose
                self.xlabel = "Dose [Gy]"  # units on x-axis
            elif target_dose > 0.0:
                _tdose = 0.001 * target_dose
                self.xlabel = "Dose [Gy]"
            else:
                _tdose = 0.1
                self.xlabel = "Dose [%]"
                self.x_is_relative = True
            logging.debug("Target dose {}".format(_tdose))
            self.x *= _tdose

            self.target_dose = target_dose

        elif cube.type == 'LET':  # LET Cube
            self.xlabel = "LET [keV/um]"

        else:  # Unknown Cube
            self.xlabel = "(unkown)"

    def write(self, filename, header=False):
        """
        Writes the DVH data to disk, using filename.
        :params str filename:
        :params bool header: select if header will be included, prefixed with #. Default: False.
        """

        with open(filename, 'w') as file:
            if header:
                file.write("# Cube basename: {}\n".format(self.cube_basename))
                if self.target_dose:
                    file.write("# Cube target dose: {} [Gy]\n".format(self.target_dose))
                file.write("# Voi name: {}\n".format(self.name))
                file.write("# X-axis: {}\n".format(self.xlabel))
                file.write("# Y-axis: {}\n".format(self.ylabel))
            for x, y in zip(self.x, self.y):
                file.write("{:.3f} {:.3f}\n".format(x, y))

    @staticmethod
    def volume_histogram(data_cube, voi=None, bins=256):
        """
        Generic volume histogram calculator, useful for DVH and LVH or similar.

        :params data_cube: a data cube of any shape, e.g. Dos.cube
        :params voi: optional voi where histogramming will happen.
        :returns [x],[y]: coordinates ready for plotting. Dose (or LET) along x, Normalized volume along y in %.

        If VOI is not given, it will calculate the histogram for the entire dose cube.

        If VOI is a point or has no contents, all y values are set to 0.

        Providing voi will slow down this function a lot, so if in a loop, it is recommended to do masking
        i.e. only provide Dos.cube[mask] instead.
        """

        if voi is None:
            mask = None
        else:
            vcube = voi.get_voi_cube()
            mask = (vcube.cube == 1000)
            if not mask.any():
                warnings.warn("Given VOI has no extend and contains no voxels.",
                              UserWarning)
                return None, None

        _xrange = (0.0, data_cube.max() * 1.1)
        _hist, x = np.histogram(data_cube[mask], bins=bins, range=_xrange)
        _fhist = _hist[::-1]  # reverse histogram, so first element is for highest dose
        _fhist = np.cumsum(_fhist)
        _hist = _fhist[::-1]  # flip back again to normal representation

        y = 100.0 * _hist / _hist[0]  # volume histograms always plot the right edge of bin, since V(D < x_pos).
        y = np.insert(y, 0, 100.0, axis=0)  # but the leading bin edge is always at V = 100.0%

        return x, y



#
#    Copyright (C) 2010-2017 PyTRiP98 Developers.
#
#    This file is part of PyTRiP98.
#
#    PyTRiP98 is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyTRiP98 is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyTRiP98.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Parser for TRiP files.
"""

import logging

from pytrip.tripexecuter import Field


logger = logging.getLogger(__name__)


class ExecParser(object):
    """
    """

    # map trip commands onto method names
    _trip_commands = {"ct": "_parse_ct",
                      # TODO: "voi": "_parse_voi",
                      "field": "_parse_field",
                      "plan": "_parse_plan",
                      "opt": "_parse_opt",
                      "optimize": "_parse_opt",
                      "scancap": "_parse_scancap"}

    # Here follows a several trip parameters/arguments which can be modified.
    # All parameters which are not supported are mapped to the _na() method,
    # which does nothing but displays a warning.

    # scancap arguments. {<trip_parameter> : (<handler_method>, <format_specifier>)}
    _scancap_args = {"offh2o": ("_update_obj", "f"),
                     "bolus": ("_update_obj", "f"),
                     "focus2stepsizefactor": ("_na", "s"),            # not implemented
                     "calbibration": ("_na", "s"),                    # not implemented
                     "path": ("_update_obj", "s", "scanpath"),
                     "rifi": ("_update_obj", "f"),
                     "couchangle": ("_na", "s"),                      # not implemented
                     "gantryangle": ("_na", "s"),                     # not implemented
                     "minparticles": ("_update_obj", "i")}

    # <attribute_name> must only be added if it is different from the name of <trip_parameter>
    # generic plan arguments. {<trip_parameter> : (<handler_method>, <format_specifier>, [<attribute_name>])}
    _plan_args = {"dose": ("_update_obj", "f", "target_dose"),
                  "targettissue": ("_update_obj", "s", "target_tissue_type"),
                  "residualtissue": ("_update_obj", "s", "res_tissue_type"),
                  "partialbiodose": ("_na", "s"),                     # not implemented
                  "incube": ("_update_obj", "s", "incube_basename"),  # TODO: may need special handler for suffix
                  "outcube": ("_na", "s"),                            # not implemented
                  "debug": ("_na", "s")}                              # not implemented

    # optimization arguments. {<trip_parameter> : (<handler_method>, <format_specifier>, [<attribute_name>])}
    _opt_args = {"iter": ("_update_obj", "i", "iterations"),
                 "graceiter": ("_na", "s"),                           # not implemented
                 "bio": ("_update_obj", "s", "opt_method"),
                 "phys": ("_update_obj", "s", "opt_method"),
                 "H2Obased": ("_update_obj", "s", "opt_principle"),
                 "CTbased": ("_update_obj", "s", "opt_principle"),
                 "singly": ("_na", "s"),                              # not implemented
                 "matchonly": ("_na", "s"),                           # not implemented
                 "dosealgorithm": ("_update_obj", "s", "dose_alg"),
                 "dosealg": ("_update_obj", "s", "dose_alg"),
                 "bioalgorithm": ("_update_obj", "s", "bio_alg"),
                 "bioalg": ("_update_obj", "s", "bio_alg"),
                 "optalgorithm": ("_update_obj", "s", "opt_alg"),
                 "optalg": ("_update_obj", "s", "opt_alg"),
                 "events": ("_na", "s"),                              # not implemented
                 "eps": ("_update_obj", "f"),
                 "geps": ("_update_obj", "f"),
                 "myfac": ("_na", "s"),                               # not implemented
                 "doseweightfactor": ("_update_obj", "f", "target_dose_percent"),  # TODO: may check 10 % or 0.1
                 "field": ("_na", "s"),                               # not implemented
                 "debug": ("_na", "s")}                               # not implemented

    # field specific arguments. {<trip_parameter> : (<handler_method>, <format_specifier>, [<attribute_name>])}
    _field_args = {"file": ("_update_obj", "s", "use_raster_file"),
                   "import": ("_na", "s"),                            # not implemented
                   "export": ("_na", "s"),                            # not implemented
                   "read": ("_na", "s"),                              # not implemented
                   "write": ("_na", "s"),                             # not implemented
                   "list": ("_na", "s"),                              # not implemented
                   "delete": ("_na", "s"),                            # not implemented
                   "display": ("_na", "s"),                           # not implemented
                   "inspect": ("_na", "s"),                           # not implemented
                   "reset": ("_na", "s"),                             # not implemented
                   "new": ("_na", "s"),                               # special case
                   "reverseorder": ("_na", "s"),                      # not implemented
                   "target": ("_na", "s"),                            # not implemented
                   "gantry": ("_update_obj", "f"),
                   "couch": ("_update_obj", "f"),
                   "chair": ("_update_obj", "f"),
                   "stereotacticcoordinates": ("_na", "s"),           # not implemented
                   "fwhm": ("_update_obj", "f"),                      # not implemented
                   "rastersteps": ("_update_obj", "[f,f]", "raster_step"),
                   "raster": ("_update_obj", "[f,f]", "raster_step"),
                   "zsteps": ("_update_obj", "f"),                    # either "zsteps" (as in TRiP98 manual)..
                   "zstep": ("_update_obj", "f", "zsteps"),           # .. or "zstep" (which is used)
                   "beam": ("_na", "s"),                              # not implemented
                   "weight": ("_na", "s"),                            # not implemented
                   "contourextension": ("_update_obj", "f", "contour_extension"),
                   "contourext": ("_update_obj", "f", "contour_extension"),
                   "doseextension": ("_update_obj", "f", "dose_extension"),
                   "doseext": ("_update_obj", "f", "dose_extension"),
                   "projectile": ("_update_obj", "s"),                # TODO: needs special handling
                   "proj": ("_na", "s"),                              # abbreviated of the above
                   "bev": ("_na", "s"),                               # not implemented
                   "nolateral": ("_na", "s"),                         # not implemented
                   "raw": ("_na", "s"),                               # not implemented
                   "dosemeanlet": ("_na", "s"),                       # not implemented
                   "algorithm": ("_na", "s"),                         # not implemented (used for .bev files only)
                   "bioalgorithm": ("_na", "s"),                      # not implemented (used for .bev files only)
                   "debug": ("_na", "s")}                             # not implemented

    def __init__(self, plan):
        self.plan = plan

    def parse_exec(self, path):
        """ Parse an .exec file and store it into self.plan

        :params str path: path to .exec file including file extension
        """
        with open(path, "r") as fp:
            data = fp.read()

        data = data.split("\n")

        for line in data:
            _key = line.split(" ")[0]  # get first word of line
            if _key in self._trip_commands:
                # lookup and call corresponding method
                logger.debug("Calling self {:s}".format(self._trip_commands[_key]))
                getattr(self, self._trip_commands[_key])(line)

    @staticmethod
    def _unpack_arg(_arg):
        """ Returns the interior of the parenthesis of an argument.

        'arg(value)'     -> 'arg', 'value'
        'arg("value")'   -> 'arg', 'value'
        'arg('value')'   -> 'arg', 'value'
        'arg'            -> 'arg', ''
        '"arg"'          -> 'arg', ''
        ''arg''          -> 'arg', ''

        Any quotes will be stripped from the value.
        Any whitespaces will be stripped from the value or arg.

        If there is no value given, simply the _arg is returned and an empty string for the value.

        example:
        >>> ExecParser._unpack_arg("bolus(2.00)")
        ('bolus', '2.00')

        """
        if "(" in _arg:
            _val = _arg[_arg.find("(") + 1:_arg.find(")")].strip("\"").strip("\'").strip()
            _arg = _arg[0:_arg.find("(")].strip()
        else:
            _val = ""
            _arg = _arg.strip("\"").strip("\'").strip()
        return _arg, _val

    def _parse_extra_args(self, _line, _dict, _obj):
        """
        Parse all args following the "/" in a given line and a given parsing dictionary.
        The parsing dictionary maps the TRiP parameter ( == argument) with a proper method.

        :params _obj: may either be self.plan or self.plan.fields[i]

        """
        items = _line.split("/")
        if len(items) > 1:
            _args = items[1].split()
            for _arg in _args:

                # there are two kinds of arguments:
                # Case 1) : foobar1(value)
                # Case 2) : foobar2

                _par, _val = self._unpack_arg(_arg)  # returns always string
                # _par holds the TRiP argument/parameter name
                # _val holds the value for the argument/parameter

                if not _val:
                    _val = _par

                # now lookup a proper method from the dict
                if _par in _dict:
                    _method = _dict[_par][0]
                    _format = _dict[_par][1]
                    # sometimes, the TRiP parameter names is not equal to the name of the attribute in _obj
                    # in these cases, a third string is in the dict, which holds the attribute name.
                    if len(_dict[_par]) > 2:
                        _objattr = _dict[_par][2]
                    else:
                        _objattr = _par

                    # _method is the name of the method to be called (key in the dicts above)
                    # _obj is the target object (the base plan or a given field)
                    # _format is the format specifier how the _val should be stored in _par for the _obj class
                    getattr(self, _method)(_obj, _objattr, _val, _format)

    def _parse_ct(self, line):
        """
        Parses the 'ct' command arguments
        """
        items = line.split("/")
        if len(items) > 1:
            path = items[0].split()[1]
            args = items[1].split()
            if "read" in args:
                self.plan.basename = path

    def _parse_plan(self, line):
        """
        Method for parsing plan data
        """
        self._parse_extra_args(line, self._plan_args, self.plan)

    def _parse_field(self, line):
        """
        Method for parsing field data
        """
        items = line.split("/")
        _number = int(items[0].split()[1])
        if len(items) > 1:
            if "new" in items[1]:
                logger.debug("Parse a new field number {:d}".format(_number))
                field = Field()
                field.number = _number
                self.plan.fields.append(field)
                self._parse_extra_args(line, self._field_args, self.plan.fields[-1])

    def _parse_scancap(self, line):
        """
        Parser for the scancap command.
        """
        self._parse_extra_args(line, self._scancap_args, self.plan)

    def _parse_opt(self, line):
        """
        Parser for the optimization command.
        """
        self._parse_extra_args(line, self._opt_args, self.plan)

    @staticmethod
    def _update_obj(_obj, _objattr, _val, _format):
        """
        :params _obj: object whose parameters will be updated
        :params _par: attribute to be set in obj
        :params _val: variable which this attribute will be set to
        """
        # Add all needed type conversions here:
        if _format == 'f':  # single float
            value = float(_val)

        elif _format == 'i':  # single integer
            value = int(_val)

        # handle tuples of arbitrary length
        # "2,2" -> (2,2) or (2.0,2.0) depending on _format = "(i,i)" or "(f,f)"
        elif "(" in _format:
            if "f" in _format:  # tuple will be float
                value = tuple(map(float, _val.split(",")))
            elif "i" in _format:  # tuple will be int
                value = tuple(map(int, _val.split(",")))

        # handle list of arbitrary length
        # "2,2" -> [2,2] or [2.0,2.0] depending on _format = "[i,i]" or "[f,f]"
        elif "[" in _format:
            if "f" in _format:  # tuple will be float
                value = list(map(float, _val.split(",")))
            elif "i" in _format:  # tuple will be int
                value = list(map(int, _val.split(",")))

        # single string (catch all)
        else:
            value = _val
        logger.debug("Setting obj.{:s}={:s}".format(_objattr, str(value)))
        setattr(_obj, _objattr, value)

    @staticmethod
    def _na(_obj, _arg1, _arg2, _format):
        """
        Not Applicable - used by not implemented TRiP98 arguments/parameters

        This method simply prints a N/A warning and exits.
        """
        logger.warning("Not implemented: '{:s}={:s} format={:s}'".format(_arg1, _arg2, _format))



#
#    Copyright (C) 2010-2017 PyTRiP98 Developers.
#
#    This file is part of PyTRiP98.
#
#    PyTRiP98 is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyTRiP98 is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyTRiP98.  If not, see <http://www.gnu.org/licenses/>.
#
"""
This example demonstrates how to load a CT cube in Voxelplan format, and the associated contours.
Then a plan is prepared and optimized using TRiP98.
"""
import os
import logging

import pytrip as pt
import pytrip.tripexecuter as pte

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)  # give some output on what is going on.

# Fist we specify the directory where all our files are:
wdir = "/home/bassler/test"

# In TRiP, the patient "TST000" would typically carry the filename "TST000000"
patient_name = "TST000000"

# so we can construc the paths to the CTX and VDX files like this:
ctx_path = os.path.join(wdir, patient_name + ".ctx")
vdx_path = os.path.join(wdir, patient_name + ".vdx")

# Next we load the CT cube:
c = pt.CtxCube()
c.read(ctx_path)

# And load the contours
v = pt.VdxCube(c)
v.read(vdx_path)

# we may print all contours found in the Vdx file, if we want to
print(v.get_voi_names())

# We need to specify where the kernel files can be found. The location may depend on the ion we
# want to treat with. This example sets up a kernel model for C-12 ions with a 3 mm Ripple Filter.
mykernel = pte.KernelModel()
mykernel.projectile = pte.Projectile("C", a=12)
mykernel.ddd_path = "/home/bassler/TRiP98/base/DATA/DDD/12C/RF3MM/*"
mykernel.spc_path = "/home/bassler/TRiP98/base/DATA/SPC/12C/RF3MM/*"
mykernel.sis_path = "/home/bassler/TRiP98/base/DATA/SIS/12C.sis"
mykernel.rifi_thickness = 3.0  # 3 mm ripple filter. (Only for documentaiton, will not affect dose optimization.)
mykernel.rifi_name = "GSI_1D_3mm"  # Additional free text for documentation.
mykernel.comment = "Carbon-12 ions with 3 mm 1D Ripple Filter"

# Ok, we have the Contours, the CT cube and dose kernels ready. Next we must prepare a plan.
# We may choose any basename for the patient. All output files will be named using
# this basename.
plan = pte.Plan(basename=patient_name, kernels=(mykernel, ))

# Plan specific data:
plan.hlut_path = "/home/bassler/TRiP98/base/DATA/HLUT/19990218.hlut"  # Hounsfield lookup table location
plan.dedx_path = "/home/bassler/TRiP98/base/DATA/DEDX/20040607.dedx"  # Stopping power tables
plan.working_dir = "/home/bassler/test/"  # working dir must exist.

# Set the plan target to the voi called "CTV"
plan.voi_target = v.get_voi_by_name('CTV')

# some optional plan specific parameters (if not set, they will all be zero by default)
plan.bolus = 0.0  # No bolus is applied here. Set this to some value, if you are to optimize very shallow tumours.
plan.offh2o = 1.873  # Some offset mimicing the monitoring ionization chambers and exit window of the beam nozzle.

# Next we need to specify at least one field, and add that field to the plan.
field = pte.Field(kernel=mykernel)  # The ion speicies is selected by passing the corresponding kernel to the field.
field.basename = patient_name  # This name will be used for output filenames, if any field specific output is saved.
field.gantry = 10.0  # degrees
field.couch = 90.0  # degrees
field.fwhm = 4.0  # spot size in [mm]

print(field)  # We can print all parameters of this field, for checking.
plan.fields.append(field)  # attach field to plan. You may attach multiple fields.

# Next, set the flags for what output should be generated, when the plan has completed.
plan.want_phys_dose = True  # We want a physical dose cube, "TST000000.dos"
plan.want_bio_dose = False  # No biological cube (Dose * RBE)
plan.want_dlet = True  # We want to have the dose-averaged LET cube
plan.want_rst = False  # Print the raster scan files (.rst) for all fields.

# print(plan)  # this will print all plan parameters

te = pte.Execute(c, v)  # get the executer object, based on the given Ctx and Vdx cube.

# in the case that TRiP98 is not installed locally, you may have to enable remote execution:
# te.remote = True
# te.servername = "titan.phys.au.dk"
# te.username = "bassler"
# te.password = "xxxxxxxx"  # you can set a password, but this is strongly discouraged. Better to exchange SSH keys!
# te.remote_base_dir = "/home/bassler/test"
#
# Depending on the remote .bashrc_profile setup, it may be needed to specify the full path
# for the remote TRiP installation. On some systems the $PATH is set, so this line can be omitted,
# or shortened to just "TRiP98" :
# te.trip_bin_path = "/opt/aptg/TRiP98/bin/TRiP98"

te.execute(plan)  # this will run TRiP
# te.execute(plan, False)  # set to False, if TRiP98 should not be executed. Good for testing.

# requested results can be found in
# plan.dosecubes[]
# and
# plan.letcubes[]
# and they are also saved to working_dir



#
#    Copyright (C) 2010-2017 PyTRiP98 Developers.
#
#    This file is part of PyTRiP98.
#
#    PyTRiP98 is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyTRiP98 is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyTRiP98.  If not, see <http://www.gnu.org/licenses/>.
#
"""
This package provides an easy way to perform interpolation of 1-D and 2-D datasets.

1-D dataset is represented by two columns of numbers: X and Y, of the same length.
X column should contain increasing sequence of numbers.
It is not required that X data should form regular (evenly spaced) grid.
X and Y columns of length 1 also form 1-D data (single data point).

2-D dataset is represented by two columns of numbers: X and Y.
and 2-dimensional matrix Z.
Number of columns in Z matrix should be equal to length of X array,
while number of rows to length of Y array.
Lengths of X and Y data may differ. One or both of X and Y columns may
contain single element - this is also valid 2-D dataset.
X and Y column should contain increasing sequence of numbers.
It is not required that X and Y data should form regular (evenly spaced) grid.

Two types of interpolation are supported: linear and spline (3rd degree polynomials).

Linear interpolation on 1-D dataset and on reduced 2-D datasets (where X or Y is single element column)
is performed using `interp` method from `numpy` package.

Spline interpolation on all datasets is performed using `InterpolatedUnivariateSpline` (for 1-D data)
and `RectBivariateSpline` (for 2-D data) from `scipy` package.

Cubic spline interpolation requires at least 4 data points. In case number of data points is lower,
interpolator implemented in this package will fall back to quadratic spline (for 3 data points)
or to linear interpolation (2 or less data points).
This is different behaviour from what is implemented in `scipy` package where exception is thrown is such situation.

Note: scipy package is not imported her as default as its installation is problematic for some group of users
(i.e. Windows users working without Anaconda python distribution).
Most of the functionality of this package (i.e. 1-D and some cases of 2-D linear interpolation) will work without
scipy package installed. If user without scipy package calls cubic spline interpolation method then an exception
will be thrown and an error message printed, requesting user to install scipy.
"""
import numpy as np
import logging

logger = logging.getLogger(__name__)


class RegularInterpolator(object):
    """
    RegularInterpolator is a helper class to easy interpolation of single- and double-variable function.
    To use it usually two steps are needed:
     1. construction of the object, providing training data (data points) and interpolation type (linear or spline).
     During this step coefficient of interpolating function will be calculated and stored in the object.
     RegularInterpolator objects are so-called callables and can be called in same way as plain functions
     on interpolated values.
     Example:

         >> interp_func_1d = RegularInterpolator(x=exp_data_x, y=exp_data_y, kind='linear')
         >> interp_func_2d = RegularInterpolator(x=exp_data_x, y=exp_data_y, z=exp_data_z, kind='spline')

    2. Calling interpolation function to get intermediate values (single number or array) on interpolated ones.

         >> interpolated_y = interp_func_1d(x=intermediate_x)
         >> interpolated_z = interp_func_2d(x=intermediate_x, y=intermediate_y)

    Interpolation in single step is also possible (although less efficient than two-step method):

         >> interpolated_y = RegularInterpolator.eval(x=intermediate_x, xp=exp_data_x, yp=exp_data_y, kind='linear')
         >> interpolated_z = RegularInterpolator(x=intermediate_x, y=intermediate_y, \
        xp=exp_data_x, yp=exp_data_y, zp=exp_data_z, kind='spline')

    """

    def __init__(self, x, y, z=None, kind='spline'):
        """
        Initialisation responsible also for finding coefficients of interpolation function.
        If z is None then we assume that dataset if 1-D and is made by X and Y columns.
        In case z is not None, then we interpret the data as 2-D.

        :param x: array_like
        The 1-d array of data-points x-coordinates, must be in strictly ascending order.

        :param y: array_like
        The 1-d array of data-points y-coordinates.
        If y is from 1-D data points (z is None) then it should be of the same length (shape) as x.
        Otherwise (2-D data points, z is not None) then it must be in strictly ascending order.

        :param z: array_like
        None in case of 1-D dataset.
        Otherwise the 2-d array of data-points z-coordinates, of shape (x.size, y.size).

        :param kind: interpolation algorithm: 'linear' or 'spline' (default).
        """

        # internal executor object
        self._interp_function = None

        # 1-D interpolation
        if z is None:
            self._interp_function = self.__get_1d_function(x, y, kind)

        # 2-D interpolation
        else:
            # 2-D data, but one of the coordinates is single-element, reducing to 1-D interpolation
            if len(x) == 1:  # data with shape 1 x N
                executor_1d = self.__get_1d_function(x=y, y=z, kind=kind)

                def applicator(x, y, *args, **kwargs):
                    return executor_1d(y)
                self._interp_function = applicator
            elif len(y) == 1:  # data with shape N x 1
                executor_1d = self.__get_1d_function(x=x, y=z, kind=kind)

                def applicator(x, y, *args, **kwargs):
                    return executor_1d(x)
                self._interp_function = applicator
            else:
                # 3-rd degree spline interpolation, passing through all points
                self._interp_function = self.__get_2d_function(x, y, z, kind)

    def __call__(self, x, y=None):
        """
        Call interpolation function on intermediate values.

        :param x: array_like or single value
        Input x-coordinate(s).
        :param y: array_like or single value
        Input y-coordinate(s) (in case of 2-D dataset).
        :return: array_like or single value
        Interpolated value
        """
        if y is None:
            return self._interp_function(x)
        else:
            return self._interp_function(x, y, grid=False)

    @classmethod
    def eval(cls, x, y=None, xp=None, yp=None, zp=None, kind='linear'):
        """
        Perform interpolation in a single step. Find interpolation function,
         based on data points (xp, yp and zp) and then execute it on interpolated values (x and y).
        :param x: array_like or single value
        Input x-coordinate(s).

        :param y: array_like or single value
        Input y-coordinate(s) (in case of 2-D dataset).

        :param xp: array_like
        The 1-d array of data-points x-coordinates, must be in strictly ascending order.

        :param yp: array_like
        The 1-d array of data-points y-coordinates.
        If y is from 1-D data points (z is None) then it should be of the same length (shape) as x.
        Otherwise (2-D data points, z is not None) then it must be in strictly ascending order.

        :param zp: array_like
        None in case of 1-D dataset.
        Otherwise the 2-d array of data-points z-coordinates, of shape (x.size, y.size).

        :param kind: interpolation algorithm: 'linear' or 'spline' (default).
        :return: array_like or single value
        Interpolated value
        """
        if xp is None or yp is None:
            logger.error("Provide valid training data points")
            raise Exception("Invalid training data points")
        interpolating_class = cls(xp, yp, zp, kind)  # find interpolating function
        return interpolating_class(x, y)  # call interpolating function

    @staticmethod
    def __get_1d_function(x, y, kind):
        """
        Train 1-D interpolator
        :param x: x-coordinates of data points
        :param y: y-coordinates of data points
        :param kind: 'linear' or 'spline' interpolation type
        :return Interpolator callable object
        """
        def fun_interp(t):
            return np.interp(t, x, y)

        # input consistency check
        if len(x) != len(y):
            logger.error("len(x) = {:d}, len(y) = {:d}. Both should be equal".format(len(x), len(y)))
            raise Exception("1-D interpolation: X and Y should have the same shape")
        # 1-element data set, return fixed value
        if len(y) == 1:
            # define constant
            def fun_const(t):
                """
                Helper function
                :param t: array-like or scalar
                :return: array of constant values if t is an array of constant scalar if t is a scalar
                """
                try:
                    result = np.ones_like(t) * y[0]  # t is an array
                except TypeError:
                    result = y[0]  # t is a scalar
                return result
            result = fun_const
        # 2-element data set, use linear interpolation from numpy
        elif len(y) == 2:
            result = fun_interp
        else:
            if kind == 'spline':
                # 3-rd degree spline interpolation, passing through all points
                try:
                    from scipy.interpolate import InterpolatedUnivariateSpline
                except ImportError as e:
                    logger.error("Please install scipy on your platform to be able to use spline-based interpolation")
                    raise e
                k = 3
                if len(y) == 3:  # fall back to 2-nd degree spline if only 3 points are present
                    k = 2
                result = InterpolatedUnivariateSpline(x, y, k=k)
            elif kind == 'linear':
                result = fun_interp
            else:
                raise ("Unsupported interpolation type {:s}.".format(kind))
        return result

    @staticmethod
    def __get_2d_function(x, y, z, kind):
        """
        Train 2-D interpolator
        :param x: x-coordinates of data points
        :param y: y-coordinates of data points
        :param z: z-coordinates of data points
        :param kind: 'linear' or 'spline' interpolation type
        :return Interpolator callable object
        """
        try:
            from scipy.interpolate import RectBivariateSpline
        except ImportError as e:
            logger.error("Please install scipy on your platform to be able to use spline-based interpolation")
            raise e
        if len(x) == 2:  # fall-back to linear interpolation
            kx = 1
        elif len(x) == 3:  # fall-back to 2nd degree spline
            kx = 2
        else:
            kx = 3
        if len(y) == 2:  # fall-back to linear interpolation
            ky = 1
        elif len(y) == 3:  # fall-back to 2nd degree spline
            ky = 2
        else:
            ky = 3
        if kind == 'linear':
            kx, ky = 1, 1
        x_array, y_array, z_array = x, y, z
        if not isinstance(x, np.ndarray):
            x_array = np.asarray(x)
        if not isinstance(y, np.ndarray):
            y_array = np.asarray(y)
        if not isinstance(z, np.ndarray):
            z_array = np.asarray(z)
        result = RectBivariateSpline(x_array, y_array, z_array, kx=kx, ky=ky, s=0)
        return result



#
#    Copyright (C) 2010-2018 PyTRiP98 Developers.
#
#    This file is part of PyTRiP98.
#
#    PyTRiP98 is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyTRiP98 is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyTRiP98.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Structure:
A Plan() may hold one or multiple Field() objects. When a Plan() is proper setup, it can be executed
with methods in Execute().
"""

import datetime
import os
import uuid
import logging

from pytrip.dos import DosCube
from pytrip.let import LETCube
from pytrip.tripexecuter.execute import Execute
from pytrip.tripexecuter.execparser import ExecParser

logger = logging.getLogger(__name__)


class Plan(object):
    """ Class of handling plans.
    """
    # dicts are in the form of
    # "<valid trip tag>": (<enum>, "Short name", "Description, e.g. for alt tags")
    opt_principles = {"H2Obased": (0, "Simple opt.", "Very simplified single field optimization"),
                      "CTbased": (1, "Full opt.", "Full optimization, multiple fields")}

    opt_methods = {"phys": (0, "Physical", "Physical dose only [Gy]"),
                   "bio": (1, "Biological", "Biological optimization [Gy(RBE)]")}

    opt_algs = {"cl": (0, "Classic", ""),
                "cg": (1, "Conjugate gradients", "(default)"),
                "gr": (2, "Plain gradients", ""),
                "bf": (3, "Bortfeld", "Bortfeld's algorithm"),
                "fr": (4, "Fletcher-Reeves", "Fletcher-Reeves' algorithm")}

    bio_algs = {"cl": (0, "Classic (default)", ""),
                "ld": (1, "Lowdose (faster)", "")}

    dose_algs = {"cl": (0, "Classic (default)", ""),
                 "ap": (1, "Allpoints", ""),
                 "ms": (2, "Multiple scatter", "")}

    scanpaths = {"none": (0, "No path", "output as is"),
                 "uw": (1, "U. Weber", "efficient"),
                 "uw2": (2, "U. Weber2", "very efficient, works also for non-grid points"),
                 "mk": (3, "M. Kraemer", "conservative")}

    def __init__(self, basename="", comment="", kernels=None):
        """
        A plan Object, which may hold several fields, general setup, and possible also output,
        if it has been calculated.
        :params str basename: TRiP98 qualified plan name. E.g. "test000001"
        :params str comment: any string for documentation purposes.
        :params kernels KernelModel:  a list of Kernel models which describes what projectile to use here.
        So far our versions of TRiP does not support multi-ion optimization.
        Therefore only a single kernel should be provided, i.e. [kernel].
        """

        self.__uuid__ = uuid.uuid4()  # for uniquely identifying this plan
        self.basename = basename.replace(" ", "_")  # TODO: also for Ctx and Vdx, issue when loading DICOMs.
        self.comment = comment

        self.fields = []  # list of Field() objects
        self.voi_target = None  # put target Voi() here. (Not needed if self.target_dose_cube is set (incube))
        self.vois_oar = []  # list of Voi() objects which are considered as OARs

        # results
        self.dosecubes = []  # list of DosCube() objects (i.e. results)
        self.letcubes = []  # list of LETCubes()

        # directories and file paths.
        self.working_dir = ""  # directory where all input files are stored, and where all output files will be put.
        self.ddd_dir = "$TRIP98/DATA/DDD/12C/RF3MM/12C.*"   # deprecated, use KernelModel
        self.spc_dir = ""  # deprecated, use KernelModel
        self.sis_path = ""  # deprecated, use KernelModel
        self.hlut_path = "$TRIP98/DATA/19990211.hlut"
        self.dedx_path = "$TRIP98/DATA/DEDX/20040607.dedx"

        self.projectile = "C"  # these are needed by makesis possibly  # deprecated, use KernelModel
        self.projectile_a = 12  # these are needed by makesis possibly  # deprecated, use KernelModel
        self.res_tissue_type = ""
        self.target_tissue_type = ""
        self.active = False
        self.optimize = True  # switch wheter it should be optimized. If False, then a RST file will be loaded instead.
        self.iterations = 500
        self.opt_method = "phys"
        self.opt_principle = "H2Obased"
        self.opt_alg = "cg"
        self.dose_alg = "cl"
        self.bio_alg = "cl"
        self.eps = 1e-3
        self.geps = 1e-4
        self.sis = None  # placeholder for sis parameters

        # Scancap parameters:
        # Thickness of ripple filter (0.0 for none)
        # will only used for the TRiP98 Scancap command.
        self.rifi = 0.0  # rifi thickness in [mm]  # deprecated, use KernelModel
        self.bolus = 0.0  # amount of bolus to be applied in [mm]
        self.offh2o = 0.0  # Offset of exit-window and ionchambers in [mm]
        self.minparticles = 5000  # smallest amount of particle which will go into a raster spot
        self.scanpath = "none"

        self.random_seed1 = 1000  # Only needed for biological optimization.
        # TODO: TRiP may accept a second "seed2" depending on PRNG. Not implemented here.

        # results requested from optimization
        self.want_phys_dose = True
        self.want_bio_dose = False
        self.want_dlet = False
        self.want_rst = False
        self.want_tlet = False  # TODO: not implemented.

        self.window = []  # window [xmin,xmax,ymin,ymax,zmin,zmax] in CTcoords [mm] for limited dose output.
        self.target_dose = 2.0  # target dose in Gray
        self.target_dose_percent = 100.0  # target dose in percent

        self.target_tissue_type = ""  # target tissue type, for biological optimization
        self.res_tissue_type = ""  # residual tissue types
        self.incube_basename = ""  # To enable incube optimization, set the basename of the cube to be loaded by TRiP

        self._trip_exec = ""   # placeholder for generated TRiP98 .exec commands.
        self._make_sis = ""  # placeholder for generate sistable command

        # TODO: for pytrip 3.0 all attributes are stored in Kernel object.
        # The API is here already prepared for multi-ion optimization, but for now it is not implemented.
        if kernels:
            kernel = kernels[0]
            if len(kernels) > 1:
                logger.warning("Multi-ion therapy is not implemented. "
                               "Using data from the first kernel in list: '{}'.".format(kernel.name))

            self.projectile = kernel.projectile.iupac
            self.projectile_a = kernel.projectile.a
            self.rifi = kernel.rifi_thickness
            self.ddd_dir = kernel.ddd_path
            self.spc_dir = kernel.spc_path
            self.sis_path = kernel.sis_path

    def __str__(self):
        """ string out handler
        """
        return self._print()

    def _print(self):
        """ Pretty print all attributes
        """
        out = "\n"
        out += "   Plan '{:s}'\n".format(self.basename)
        out += "----------------------------------------------------------------------------\n"
        out += "|  UUID                         : {:s}\n".format(str(self.__uuid__))
        if self.voi_target:
            out += "|  Target VOI                   : {:s}\n".format(self.voi_target.name)
        else:
            out += "|  Target VOI                   : (none set)\n"
        if self.vois_oar:
            for _oar in self.vois_oar:
                out += "|  Organs at Risk VOIs          : {:s}\n".format(_oar.name)
        else:
            out += "|  Organs at Risk VOIs          : (none set)\n"

        if self.fields:
            out += "+---Fields\n"
            for _field in self.fields:
                out += "|   |           #{:d}              : '{:s}'\n".format(_field.number, _field.basename)
        else:
            out += "|   Fields                      : (none set)\n"

        if self.dosecubes:
            out += "+---Dose cubes:\n"
            for _dosecube in self.dosecubes:
                out += "|   |                              : {:s}\n".format(_dosecube.name)
        else:
            out += "|  Dose cubes                   : (none set)\n"

        if self.letcubes:
            out += "+---LET cubes:\n"
            for _letcube in self.letcubes:
                out += "|   |                              : {:s}\n".format(_letcube.name)
        else:
            out += "|  LET cubes                    : (none set)\n"

        out += "|\n"
        out += "| Directories\n"
        out += "|   Working directory           : {:s}\n".format(self.working_dir)
        out += "|   DDD directory               : {:s}\n".format(self.ddd_dir)
        out += "|   SPC directory               : {:s}\n".format(self.spc_dir)
        out += "|   SIS path                    : {:s}\n".format(self.sis_path)
        out += "|   HLUT path                   : {:s}\n".format(self.hlut_path)
        out += "|   dE/dx path                  : {:s}\n".format(self.dedx_path)

        out += "|\n"
        out += "| Sis table generation\n"
        out += "|   {:s}\n".format(self._make_sis)

        out += "|\n"
        out += "| Optimization parameters\n"
        out += "|   Optimization enabled        : {:s}\n".format(str(self.optimize))
        out += "|   Optimization method         : '{:s}' {:s}\n".format(self.opt_method,
                                                                        self.opt_methods[self.opt_method][1])
        out += "|   Optimization principle      : '{:s}' {:s}\n".format(self.opt_principle,
                                                                        self.opt_principles[self.opt_principle][1])
        out += "|   Optimization algorithm      : '{:s}' {:s}\n".format(self.opt_alg,
                                                                        self.opt_algs[self.opt_alg][1])
        out += "|   Dose algorithm              : '{:s}' {:s}\n".format(self.dose_alg,
                                                                        self.dose_algs[self.dose_alg][1])
        out += "|   Biological algorithm        : '{:s}' {:s}\n".format(self.bio_alg,
                                                                        self.bio_algs[self.bio_alg][1])
        out += "|   Iterations                  : {:d}\n".format(self.iterations)
        out += "|   eps                         : {:.2e}\n".format(self.eps)
        out += "|   geps                        : {:.2e}\n".format(self.geps)

        out += "|\n"
        out += "| Scanner capabilities (Scancap)\n"
        out += "|   Ripple filter thickness     : {:.3f} [mm]\n".format(self.rifi)
        out += "|   Bolus thickness             : {:.3f} [mm]\n".format(self.bolus)
        out += "|   H2O offset                  : {:.3f} [mm]\n".format(self.offh2o)
        out += "|   Min particles               : {:d}\n".format(self.minparticles)
        out += "|   Scanpath                    : '{:s}'\n".format(self.scanpath,
                                                                   self.scanpaths[self.scanpath])
        out += "|\n"
        out += "| Optimization target\n"
        out += "|   Relative target dose        : {:.1f} %\n".format(self.target_dose_percent)
        out += "|   100.0 % target dose set to  : {:.2f} [Gy]\n".format(self.target_dose)
        if self.incube_basename:
            out += "|   Incube optimization         : {:s}\n".format(self.incube_basename + '.dos')
        else:
            out += "|   Incube optimization         : (none set)\n"

        out += "|\n"
        out += "| Biological parameters\n"
        out += "|   Target tissue type          : '{:s}'\n".format(self.target_tissue_type)
        out += "|   Residual tissue type        : '{:s}'\n".format(self.res_tissue_type)
        out += "|   Random seed #1              : {:d}\n".format(self.random_seed1)

        out += "|\n"
        out += "| Requested output\n"
        out += "|   Physical dose cube          : {:s}\n".format(str(self.want_phys_dose))
        out += "|   Biological dose cube        : {:s}\n".format(str(self.want_bio_dose))
        out += "|   Dose averaged LET cube      : {:s}\n".format(str(self.want_dlet))
        out += "|   Track averaged LET cube     : {:s}\n".format(str(self.want_tlet))
        out += "|   Raster scan files           : {:s}\n".format(str(self.want_rst))
        if self.window:
            out += "|   Cube output window\n"
            out += "|      Xmin / Xmax              : {:.2f}\n".format(self.window[0], self.window[1])
            out += "|      Ymin / Ymax              : {:.2f}\n".format(self.window[2], self.window[3])
            out += "|      Zmin / Zmax              : {:.2f}\n".format(self.window[4], self.window[5])
        else:
            out += "|   Cube output window          : (none set)\n"
        return out

    def save_plan(self, images, path):
        """ Saves the complete plan to path.
        1) *.exec
        2) *.dos
        """
        self.save_exec(images, path + ".exec")
        self.save_data(images, path + ".exec")

    def save_exec(self, exec_path=None):
        """ Generates (overwriting) self._trip_exec, and saves it to exec_path.
        """

        # _trip_exec is marked, private, users should not touch it.
        # Here it will be overwritten no matter what.
        self.make_exec()

        if exec_path is None:
            exec_path = os.path.join(self._temp_dir, self.basename, "*.exec")

        with open(exec_path, "w") as f:
            f.write(self._trip_exec)

    def read_exec(self, exec_path):
        """ Reads an .exec file onto self.
        """
        _exec = ExecParser(self)
        _exec.parse_exec(exec_path)

    def save_data(self, images, path):
        """ Save this plan, including associated data.
        TODO: may have to be implemented in a better way.
        """
        t = Execute(images)
        t.set_plan(self)
        t.save_data(path)
        for dos in self.dosecubes:
            dos.write(path + "." + dos.get_type() + DosCube.data_file_extension)

        for let in self.letcubes:
            let.write(path + "." + dos.get_type() + DosCube.data_file_extension)

    def load_let(self, path):
        """ Load and append a new LET cube from path to self.letcubes.
        """

        let = LETCube()
        let.read(os.path.splitext(path)[0] + ".dos")

        # TODO: check if cube is already loaded, if so, dont load.
        # UUID for LET cubes would be useful then?
        self.letcubes.append(let)

    def clean_voi_caches(self):
        """ TODO: document me
        """
        for voi in self.vois:
            voi.clean_cache()

    def load_dose(self, path, _type, target_dose=0.0):
        """ Load and append a new DosCube from path to self.doscubes.
        """
        dos = DosCube()
        dos.read(os.path.splitext(path)[0] + DosCube.data_file_extension)
        dos._type = _type
        dos.target_dose = target_dose
        self.dosecubes.append(dos)

    def destroy(self):
        """ Destructor for Vois and Fields in this class.
        """
        self.vois.destroy()
        self.fields.destroy()

    def make_exec(self, no_output=False):
        """
        Generates the .exec script from the plan and stores it into self._trip_exec.
        """

        logger.info("Generating the trip_exec script...")

        # All plans must be the same projectile
        # TODO: throw some error if this is not the case
        # projectile = self.fields[0].projectile

        output = []

        output.extend(self._make_exec_header())  # directories and scancap
        output.extend(self._make_exec_input())  # input data, CT and VDX
        output.extend(self._make_exec_fields())  # setup all fields
        output.extend(self._make_exec_oars())  # define OARs

        # for incube optimization
        if self.incube_basename:
            logger.info("Incube optimization selected by {:s}.dos".format(self.incube_basename))
            output.extend(self._make_exec_plan(incube=self.incube_basename))
        else:
            output.extend(self._make_exec_plan())

        # optimization is optional, alternatively do a forward calculation based on .rst file, or do nothing.
        if self.optimize:
            output.extend(self._make_exec_opt())

        # attach the wanted output:
        output.extend(self._make_exec_output(self.basename, self.fields))

        output.extend(["exit"])
        out = "\n".join(output) + "\n"
        self._trip_exec = out
        return out

    def _make_exec_header(self):
        """
        Add some intro header and timestamp.
        Prepare sis, hlut, ddd, dedx, and scancap commands.
        # TODO: scancap could go out into a new method.
        # TODO: check input params
        :returns: an array of lines, one line per command.
        """

        output = []
        output.append("* {:s} created by PyTRiP98 on the {:s}".format(self.basename + ".exec",
                                                                      str(datetime.datetime.now())))
        # TODO: add user and host
        # We can only check if dir exists, if this is supposed to run locally.

        if hasattr(self, "UUID"):
            output.append("*$$$ Plan UUID {:s}".format(self.UUID))
        output.append("*")
        output.append("time / on")
        output.append("sis  * /delete")
        output.append("hlut * /delete")
        output.append("ddd  * /delete")
        output.append("dedx * /delete")
        output.append('dedx "{:s}" / read'.format(self.dedx_path))
        output.append('hlut "{:s}" / read'.format(self.hlut_path))

        # ddd, spc, and sis:
        output.append('ddd "{:s}" / read'.format(self.ddd_dir))

        if self.spc_dir:  # False for None and empty string.
            output.append('spc "{:s}" / read'.format(self.spc_dir))

        if self.sis_path:
            output.append('sis "{:s}" / read'.format(self.sis_path))
        else:
            if not self._make_sis:
                logger.error("No SIS table loaded or generated.")
                raise
            output.append(self._make_sis)

        # Scancap:
        opt = "scancap / offh2o({:.3f})".format(self.offh2o)
        opt += " rifi({:.3f})".format(self.rifi)
        opt += " bolus({:.3f})".format(self.bolus)
        opt += " minparticles({:d})".format(self.minparticles)
        opt += " path({:s})".format(self.scanpath)
        output.append(opt)

        output.append("random {:d}".format(self.random_seed1))
        return output

    def _make_exec_input(self):
        """ Add CT and target VOI to exec file.
        """

        _name = self.voi_target.name.replace(" ", "_")

        output = []
        output.append("ct \"{:s}\" / read".format(self.basename))
        output.append("voi \"{:s}\" / read select(\"{:s}\")".format(self.basename,
                                                                    _name))
        output.append("voi \"{:s}\" / maxdosefraction({:.3f})".format(_name,
                                                                      self.target_dose_percent * 0.01))
        return output

    def _make_exec_fields(self):
        """ Generate .exec command string for one or more fields.
        if Plan().optimize is False, then it is expected there will
        be a path to a rasterscan file in field.rst_path

        :returns: an array of lines, one line per field in fields.
        """
        fields = self.fields

        output = []
        for i, _field in enumerate(fields):
            if self.optimize:
                # this is a new optimization:
                line = "field {:d} / new".format(i + 1)
            else:
                # or, there is a precalculated raster scan file which will be used instead:
                line = "field {:d} / read file({:s})".format(i + 1, _field.rasterfile_path)

            line += " fwhm({:.3f})".format(_field.fwhm)

            line += " raster({:.2f},{:.2f})".format(_field.raster_step[0],
                                                    _field.raster_step[1])
            # TODO: convert if Dicom angles were given
            # gantry, couch = angles_to_trip(_field.gantry(), _field.couch())
            line += " couch({:.1f})".format(_field.couch)
            line += " gantry({:.1f})".format(_field.gantry)

            # set isocenter if specified in field
            _ic = _field.isocenter
            if len(_ic) >= 3:
                line += " target({:.1f},{:.1f},{:.1f}) ".format(_ic[0], _ic[1], _ic[2])

            # set dose extension:
            # TODO: check number of decimals which make sense
            # TODO: zeros allowed?
            line += " doseext({:.4f})".format(_field.dose_extension)
            line += " contourext({:.2f})".format(_field.contour_extension)
            line += " zsteps({:.3f})".format(_field.zsteps)
            line += ' proj({:s})'.format(_field.kernel.projectile.iupac)  # TODO: check if 'C' or better '12C"
            output.append(line)
        return output

    def _make_exec_oars(self):
        """ Generates the list of TRiP commands specifying the organs at risk.
        :params [str] oar_list: list of VOIs which are organs at risk (OARs)
        """

        output = []
        for oar in self.vois_oar:
            _out = "voi " + oar.name.replace(" ", "_")
            _out += " / maxdosefraction({.3f}) oarset".format(oar.max_dose_fraction)
            output.append(_out)
        return output

    def _make_exec_plan(self, incube=""):
        """ Generates the plan command in TRiP98
        """
        output = []
        plan = "plan / dose({:.2f}) ".format(self.target_dose)
        if self.target_tissue_type:
            plan += "target({:s}) ".format(self.target_tissue_type)
        if self.res_tissue_type:
            plan += "residual({:s}) ".format(self.res_tissue_type)
        if not incube == "":
            plan += "incube({:s})".format(self.incube_basename)
        output.append(plan)
        return output

    def _make_exec_opt(self):
        """ Generates the optimizer command for TRiP based on Plan() loaded into self.
        :returns: an array of lines, but only holding a single line containing the entire command.
        """
        # example output:
        # opt / field(*) ctbased bio dosealg(ap) optalg(cg) bioalg(ld) geps(1E-4) eps(1e-3) iter(500)
        #

        opt = "opt / field(*)"

        if self.opt_principle not in self.opt_principles:
            logger.error("Unknown optimization principle {:s}".format(self.plan.opt_principle))
        opt += " {:s}".format(self.opt_principle)  # ctbased or H2Obased

        if self.opt_method not in self.opt_methods:
            logger.error("Unknown optimization method {:s}".format(self.plan.opt_method))
        opt += " {:s}".format(self.opt_method)  # "phys" or "bio"

        if self.dose_alg not in self.dose_algs:
            logger.error("Unknown optimization dose algorithm{:s}".format(self.plan.dose_alg))
        opt += " dosealg({:s})".format(self.dose_alg)  # "ap",..

        if self.opt_alg not in self.opt_algs:
            logger.error("Unknown optimization method {:s}".format(self.plan.opt_alg))
        opt += " optalg({:s})".format(self.opt_alg)  # "cl"...

        # TODO: sanity check numbers
        opt += " geps({:.2e})".format(self.geps)
        opt += " eps({:.2e})".format(self.eps)
        opt += " iter({:d})".format(self.iterations)

        return [opt]

    def _make_exec_output(self, basename, fields):
        """
        Generate TRiP98 exec commands for producing various output.
        :params str basename: basename for output files (i.e. no suffix) and no trailing dot
        :params fields:

        :returns: output string
        """
        output = []
        window = self.window
        window_str = ""
        if len(window) == 6:
            window_str = " window({:.2f},{:.2f},{:.2f},{:.2f},{:.2f},{:.2f} ".format(window[0], window[1],  # Xmin/max
                                                                                     window[2], window[3],  # Ymin/max
                                                                                     window[4], window[5])  # Zmin/max

        self._out_files = []  # list of files generated which will be returned

        if self.want_phys_dose:
            line = 'dose "{:s}."'.format(basename)
            self._out_files.append(basename + ".phys.dos")
            self._out_files.append(basename + ".phys.hed")
            line += ' /calculate  alg({:s})'.format(self.dose_alg)
            line += window_str
            line += ' field(*) write'
            output.append(line)

        if self.want_bio_dose:
            line = 'dose "{:s}."'.format(basename)
            self._out_files.append(basename + ".bio.dos")
            self._out_files.append(basename + ".bio.hed")
            line += ' /calculate  bioalgorithm({:s})'.format(self.bio_alg)
            line += window_str
            line += ' biological norbe field(*) write'
            output.append(line)

        if self.want_dlet:
            line = 'dose "{:s}."'.format(basename)
            self._out_files.append(basename + ".dosemlet.dos")
            self._out_files.append(basename + ".dosemlet.hed")
            line += ' /calculate  alg({:s})'.format(self.dose_alg)
            line += window_str
            line += ' field(*) dosemeanlet write'
            output.append(line)

        if self.want_rst and self.optimize:
            for i, field in enumerate(fields):
                output.append('field {:d} / write file({:s}.rst) reverseorder '.format(i + 1, field.basename))
                self._out_files.append(field.basename + ".rst")

        # TODO: add various .gd files
        return output

    def _make_exec_inp_data(self, projectile=None):
        """ Prepare CTX, VOI loading and RBE base data

        :returns: an array of lines, one line per command.
        """
        output = []

        output.append("ct {:s} / read".format(self._plan_name))  # loads CTX
        output.append("voi {:s} / read".format(self._plan_name))  # loads VDX
        output.append("voi * /list")  # display all loaded VOIs. Helps for debugging.

        # TODO: consider moving RBE file to DDD/SPC/SIS/RBE set
        # TODO: check rbe.get_rbe_by_name method
        if self.target_tissue_type:  # if not None or empty string:
            rbe = self.rbe.get_rbe_by_name(self.target_tissue_type)
            output.append("rbe '{:s}' / read".format(rbe.path))

        if self.res_tissue_type:
            rbe = self.rbe.get_rbe_by_name(self.target_tissue_type)
            output.append("rbe '{:s}' / read".format(rbe.path))

        return output

    def make_sis(self, projectile, focus=(), intensity=(), position=(), path="", write=False):
        """
        Creates a SIS table based on path, focus, intensity and position.
        example:
        plan.make_sis("1H", "4", "1E6,1E7,1E8", (2, 20, 0.1))
        results in:
        makesis 1H / focus(4) intensity(1E6,1E7,1E8) position(2 TO 20 BY 0.1)

        :params str projectile" such as "12C" or "1H"
        :params str focus: comma-delimited list of FWHM focus positions in [mm], such as "2,4,6,8"
        :params str intensities: tuple/list of intensities, such as "1E6,1E7,1E8"
        :params float position: tuple describing min/max/step range, i.e. (5,40,0.1) for 5 to 40 cm in 1 mm steps.
        :params path: write path
        :params write: Set to True, if the generated sis file should be written to path when executed.
        """

        self._make_sis = "makesis {:s} /".format(projectile)
        if focus:
            self._make_sis += " focus({:s})".format(focus)
        if intensity:
            self._make_sis += " intensity({:s})".format(intensity)
        if position:
            self._make_sis += " position({:.2f} TO {:.2f} BY {:.2f})".format(position[0], position[1], position[2])

        if write:
            self._make_sis += "sis \n{:s} / write".format(path)



#
#    Copyright (C) 2010-2017 PyTRiP98 Developers.
#
#    This file is part of PyTRiP98.
#
#    PyTRiP98 is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyTRiP98 is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyTRiP98.  If not, see <http://www.gnu.org/licenses/>.
#
"""
TCP models
"""

import numpy as np
import logging

logger = logging.getLogger(__name__)


def tcp_voi(sf, voi=None, ncells=1.0, fractions=1):
    """
    Returns TCP within VOI.
    If VOI is not give, TCP of entire cube is calculated.
    This is equation (7) in https://doi.org/10.1093/jrr/rru020
    assuming static oxygenation during all fractions.
    (Equation (8) would require a new oxy cube after every fractionation, not implemented.)

    :params numpy.array sf: numpy array, surviving fraction cube
    :params Voi voi: pytrip Voi() class object
    :params float ncells: number of cells in each voxel, or a cube of surviving fractions
    :params int fractions: number of fractions, default is 1
    """

    tcp = 0

    if voi is None:
        mask = None
    else:
        # mark ony those values as true, which are within the VOI
        voi_cube = voi.get_voi_cube()
        mask = (voi_cube.cube == 1000)

    # ncells may be either a scalar or a cube.
    if np.isscalar(ncells):
        tcp = np.exp(-sum(ncells * sf[mask]**fractions))
    else:
        # better make sure that the cells cube has the same size as the surviving fraction cube
        if ncells.shape == sf.shape:
            tcp = np.exp(-sum(ncells[mask] * sf[mask]**fractions))
        else:
            logger.error("ncells array shape does not match surviving fraction shape.")

    return tcp
